################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################
require File.join( File.dirname(__FILE__), "util")


################################################################################
# Globals
################################################################################

  ################################################################################
  # Constants
  ################################################################################
  
  # HTTP Header Names
  # =================
  HTTP_ACCEPT = "Accept"
  HTTP_CONTENT_TYPE = "Content-Type"
  HTTP_ACCEPT_LANGUAGE = "Accept-Language"
  
  # HTTP Header Values
  # ==================
  APPLICATION_JSON = "application/json"
  LANGUAGE_EN = "en"
  
################################################################################
# Class Definition
################################################################################

module DevopsDeploymentLib
  
  # DevopsServerInterface provides an interface / client for 
  # interacting with the DevopsServer
  class DevopsServerInterface
    include DevopsDeploymentLib::Util
    
    attr_accessor :bindings
    
    attr_reader :resource_url
    attr_reader :view_url
    attr_reader :id
    attr_reader :cloud_resource_view_url

    # Initialize
    def initialize(env_def_id, devops_service_url, rtc_url, iwd_url, log)
      @devops_service_url = devops_service_url
      @rtc_url = rtc_url
      @iwd_url = iwd_url
      @log = log
      @env_def_id = env_def_id
      @payload = nil
      @http_client = HTTPClient.new
    
      if @devops_service_url
        @resource_url = "#{@devops_service_url}/environments/#{@env_def_id}"
        @view_url = "#{@devops_service_url}/index.html?environmentID=/environments/#{@env_def_id}"
      else
        @resource_url = nil
        @view_url = nil
      end
    end
    

    # Post
    # Notify the DevOps server that a build has happened including (build id, build status, url of the build in RTC) 
    # and that an infrastructure vm is being started including (vm status, url of vm in IWD)
    def post(build_id, build_uuid, application_status, infrastructure_status)
      
      # VSP ID
      # APPLICATION_URL 
      # VIEW URL 
      # BUILD_ID
      # BUILD_UUID
      # APPLICATION_STATUS
      # INFRASTRUCTURE_STATUS
      # CLOUD_SERVICE_RESPONSE
      # RTC URL ( PROPS['rtc_properties']['url'] ) 
      
      # TODO: vsp_id should not be hard coded to "1684"
      vsp_id = "1684"
      
      response = nil
      virtual_system_id = nil
      
      @payload = nil
      
      if @devops_service_url

        if posted? 
          @log.info "[DevOps Service]: A redundant POST request was attempted for environment #{@id}."
          return false
        end
        
        # REMOVED TO rakefile build_system.publish_link(build, view_env_url, "DevOps Environment (ID: #{env_def_id})")
        
        # assumes log is a global
        @log.debug "[DevOps Service]:  Environment resource URL: #{@resource_url}"
        
        body = {
          "label" => build_id,
          "application_build" => [ 
            {
              "id" => build_id,
              "status" => application_status,
              "url" => "#{@rtc_url}/resource/itemOid/com.ibm.team.build.BuildResult/#{build_uuid}"
            }
          ], 
          "infrastructure" => [
            {
              "status" => infrastructure_status,
              "url" => "#{@iwd_url}/dashboard/runtime/virtual/##{vsp_id}"
            }
          ]    
        }
          
        args = {
          :body => body.to_json,
          :header => {
            HTTP_CONTENT_TYPE => APPLICATION_JSON
          }
        }
        
        @log.debug "[DevOps Service]: POST Request #{body.to_json}"
        
        # disable as this can fail the whole build without a valid server
        begin
          response = @http_client.post(@resource_url, :body => args[:body], :header => args[:header])
          @log.debug "[DevOps Service]: POST Response: #{response.to_json}" 
          @payload = JSON.parse(response.body)
          @id = @payload['id']
        rescue 
          @log.info "[DevOps Service]: Could not post information about DevOps environment to #{@resource_url}. Received Response #{response}."
          return false
        end
        return @id.nil? ? false : true # successful communication with DevOps Service
      end
      
      return false # no devops_service_url provided 

    end
    
    
    # Posted?
    # Answer true if a post has happened
    def posted? 
      return !@id.nil? && !@payload.nil?
    end
    
    
    # Update
    # Update the application build status, the infrastructure status and url of the vm in IWD
    def update (cloud_resource_id, application_status, infrastructure_status)
       
      if @devops_service_url && posted?
         
        if cloud_resource_id
          @cloud_resource_view_url = "#{@iwd_url}/dashboard/runtime/virtual/##{cloud_resource_id}"
          @payload['infrastructure'][0]['url'] = @cloud_resource_view_url
        end
         
        if application_status
          @payload['application_build'][0]['status'] = application_status
        end
        
        if infrastructure_status
          @payload['infrastructure'][0]['status'] = infrastructure_status
        end
        
        args = {
          :body => @payload.to_json,
          :header => {
            HTTP_CONTENT_TYPE => APPLICATION_JSON
          }
        }
        # disable as this can fail the whole build without a valid server
        begin
          @log.debug "[DevOps Service]: Updating environment #{@id} at #{@resource_url}/deployments/#{@id}"
          @http_client.put("#{@devops_service_url}/deployments/#{@id}", :body => args[:body], :header => args[:header])
        rescue
          @log.info "[DevOps Service]: Could not post information about DevOps environment to #{@resource_url}/deployments/#{@id}"
        end
      end      
      
    end    
    

  end
end
