require 'json'
require 'weaver/core/logging'
require 'weaver/utils/value_decrypter'
require 'weaver/dsl/weaver_parser'
require 'weaver/deployers/execution/continuous_validation_setup'

module Weaver
  module Chef
    
    # Author:: Florian Rosenberg
    class ChefAutomationHandler
      include Weaver::Core
      
      # Constructor
      def initialize(&reporting)
        @keyfile = nil
        @reporting = reporting 
      end
      
      # Inject the deps_file_url into each "chef_automation" statement in the folded environment topology.
      # * *Args*    :
      #   -+environment+ a topology (Environment or Topology instance)
      #   -+deps_file_url+ The dependency file URL containing a link that describes what Chef recipes to fetch (String)
      # * *Returns*
      #    +Environment+ the updated environment or topology
      def inject_deps_file_url(environment, deps_file_url) 
        raise WeaverError, "Type error for 'environment': Environment or Topology instance expected." unless environment.kind_of?(Environment) || environment.kind_of?(Topology)
        # we accept both environment or topology
        topology = nil
        if environment.kind_of?(Environment)
          topology = environment.topology
        else
          topology = environment
        end
        
        unless topology.nil?
          topology.flatten_children.select { |a| a.kind_of?(Automation) && a.type == :chef_automation }.each do |automation|
            if automation.amd[:deps_file_url] && automation.amd[:deps_file_url].implicit? 
              automation.deps_file_url = deps_file_url
              automation.amd[:deps_file_url].implicit = false
            elsif automation.amd[:deps_file_url].nil? 
              @@logger.warn "#{automation.fully_qualified_id} does not have a 'deps_file_url' attribute."
            elsif automation.deps_file_url != deps_file_url
              @@logger.warn "Cannot set #{automation.fully_qualified_id} to #{deps_file_url.inspect} because it was previously set to #{automation.amd[:deps_file_url].value.inspect}"
            end            
          end
        end
        
        return topology
      end
      
      # Generates a set of Chef nodefiles that are required for the Chef solo bootstrapping.
      # The node files will be named according to the node ids in the topology plus a suffix "_node.json".
      # * *Args*    :
      #   -+environment+ the environment topology that contains "chef_automation" statements (Environment or Topology instance)
      #   -+directory+ a directory where the files are being stored
      #
      def generate_node_files(environment, directory)        
        # We need to create a structure like this - Chef node files in JSON format.
        #{
        #  "run_list": [ "role[jke_database_role]", "role[other_role]" ]
        #}
        
        raise WeaverError, "Type error for 'environment': Environment or Topology instance expected." unless environment.kind_of?(Environment) || environment.kind_of?(Topology)
        
        # we accept both environment or topology
        topology = nil
        if environment.kind_of?(Environment) 
          topology = environment.topology
        else
          topology = environment 
        end
        
        generated_files = []
        topology.flatten_children.select { |n| n.respond_to?(:realized_by) && n.realized_by.empty? }.each do |node|
          node_json = get_node_runlist(node)

          # write to output file
          file = File.join(directory, "#{node.id.to_s}_node.json")
          @reporting.call(Logger::INFO, "Node file '#{file}'") if @reporting
          File.open(file, "w") do |f|
            f.write JSON.pretty_generate(node_json)+"\n"             
          end
          generated_files << file
        end
        return generated_files        
      end
      
      def generate_node_file(node, directory, automation_filter = [])        
        # We need to create a structure like this - Chef node files in JSON format.
        #{
        #  "run_list": [ "role[jke_database_role]", "role[other_role]" ]
        #}
        
        raise WeaverError, "Type error for 'node': Node expected." unless node.kind_of?(Environment) || environment.kind_of?(Weaver::Core::Node)
        
        node_json = get_node_runlist(node)

        # write to output file
        file = File.join(directory, "#{node.id.to_s}_node.json")
        @reporting.call(Logger::INFO, "Node file '#{file}'") if @reporting
        File.open(file, "w") do |f|
          f.write JSON.pretty_generate(node_json)+"\n"             
        end

        return file        
      end

      # Get a Hash of the run_list of the node, based on the chef automations in the node
      # The Hash will have one entry with the key "run_list" pointing to an array of role strings
      # * *Args*    :
      #   -+node+ the node that contains chef automations
      #   -+automation_filter+ list of automations on the node that are desired; empty implies all
      #
      def get_node_runlist(node, automation_filter = [])
        run_list = []
        node_json = { "run_list" => run_list }
        
        all_automation = node.flatten_children.select { |a| a.kind_of?(Automation) && a.type == :chef_automation }
        automation_order = node.automation_order.select{ |a| a.kind_of?(Automation) && a.type == :chef_automation }
        
        # process chef_automation from the automation order first
        automation_order.each do |a|
          run_list << "role[#{a.role}]" if satisfies_automation_filter(a, automation_filter)
        end
       
        # process all remaining chef_automations that are not part of the automation_order        
        all_automation.each do |a|
          run_list << "role[#{a.role}]" unless automation_order.include?(a) or !satisfies_automation_filter(a, automation_filter)         
        end
        return node_json
      end
            
      # In each script package, if chef_node_attr is a Hash, convert it to a String.
      # If one of the hash values is an interpolation, the interpolation will be evaluated against the infrastructure
      # The common case where chef_node_attr values are specified directly on a script package is when an infrastructure is
      # used in conjunction with a yaml file that holds its exported properties.
      # * *Args*   
      #   -+infrastructure+ the infrastructure whose chef_node_attrs are being converted
      #   -+node+ a node from the topology that contains "chef_automation" statements 
      def interpolate_and_convert_chef_attributes(infrastructure, node)
        node.automation.select { |k, v|
          v.type == :script_package && v.amd[:chef_node_attr] }.each do |k, v|
          @@logger.debug "Interpolating and converting Chef attributes on #{v.fully_qualified_id}"
          v.amd[:chef_node_attr].value = self.class.interpolate_and_convert_chef_node_attr_to_s(v.amd[:chef_node_attr].value, Sandbox.new(infrastructure, nil, '', {}, {}, AllCache.new))
          @@logger.debug "Interpolated and converted value of #{v.fully_qualified_id} is #{v.amd[:chef_node_attr].value}"
          # if there is a chef_node, chef_node_attr should not be left with a null or empty string; it should be set to a blank string
          if v.amd[:chef_node]
            if v.amd[:chef_node_attr].value.nil? || v.amd[:chef_node_attr].value.empty?
              v.amd[:chef_node_attr].value = " "
            end
          end
        end
      end
      
      # Generates the Chef related attributes that the DevOps bootstrapping mechanism uses from all chef automations.
      # CHEF_NODE, CHEF_NODE_ATTR and DEPS_FILE_URL.
      # * *Args*    :
      #   -+node+ a node from the topology that contains "chef_automation" statements
      #   -+automation_filter+ list of automations on the node that are desired; empty implies all
      # * *Returns*
      #    +Hash+ a hash containing three elements (keys are symbols): 
      #         +chef_node+ with the name of the node file that the DevOps bootstrap needs
      #         +deps_file_url+ a URL pointing to a file that contains which cookbooks to fetch
      #         +chef_node_attr+ a Hash that tell Chef what attributes it need to pass it to the Chef solo run  
      def generate_chef_attributes(node, automation_filter = [])
        deps_file_url = nil
        chef_node_attr = map = {}
        obscured_chef_node_attr = obscured_map = {}

        node.flatten_children.select { |a| a.kind_of?(Automation) && a.type == :chef_automation && satisfies_automation_filter(a, automation_filter)}.each do |automation|
          # get the deps_file_url - more than one different deps_file_url is currently not supported.
          if !automation.deps_file_url.nil? && !deps_file_url.nil? && (deps_file_url != automation.deps_file_url)
            @@logger.warn "More than one non-unique 'deps_file_url' attribute values found. This is currently not supported."
          else
            deps_file_url = automation.deps_file_url
          end

          # Build up chef_node_attr (TODO is is not clear how the chef string has to look like if complex 
          # attribute structures are used in default.rb). Currently this is the way attributes in chef are specified.
          # default['jke']['attribute1'] = 'foo'                   
          # default['jke']['attribute1']['x'] = 'bar'.
          # 
          # It is not clear whether the above is allowed in Chef nor how the JSON to structure that is used
          # in Chef solo has to look like. Currently, we assume that it is the following:  
          # { 'jke' => { 'attribute1' => [ 'foo' , { 'x' => 'bar']
          #
          # (Note we now also build up obscured_chef_node_attr using the exact same technique.)
                  
          automation.amd.select { |k,v| v.mutable && !v.implicit? && !v.name.nil? }.each do |id, p|
            split = p.name.split(".") # split up the attribute string for chef            
            split.each_with_index do |elem, i| 
              if i == split.length-1 # if we hit the last one -> actual value
                newval = p.value.respond_to?(:ciphertext) ? Weaver::Utils::ValueDecryptor.decrypt(p.value, @keyfile) : p.value
                if map.key?(elem) 
                  val = map[elem]
                  map[elem] = [val, newval] unless val == newval
                  obscured_map[elem] = [obscured_map[elem], p.value] unless obscured_map[elem] == p.value 
                else 
                  map[elem] = newval
                  obscured_map[elem] = p.value
                end
              else
                map[elem] = {} unless map.key?(elem)
                obscured_map[elem] = {} unless obscured_map.key?(elem)
              end
              map = map[elem]
              obscured_map = obscured_map[elem]
            end
            map = chef_node_attr
            obscured_map = obscured_chef_node_attr
          end
        end
        #@logger.debug chef_node_attr.inspect
        weaver_automation = chef_node_attr['weaver']
        unless weaver_automation.nil? 
           continuous_validation = chef_node_attr['weaver']['continuous_validation']
           Weaver::Deployers::Execution::ContinuousValidationSetup.format_chef_node_attr(chef_node_attr)  unless continuous_validation.nil?           
        end
        
        return { :chef_node => node.id.to_s, :chef_node_attr => self.class.convert_chef_node_attr_to_s(chef_node_attr), :deps_file_url => deps_file_url, :obscured_chef_node_attr => self.class.convert_chef_node_attr_to_s(obscured_chef_node_attr)}
      end
      
      def has_chef_automation(node)
        chef_automations = node.flatten_children.select { |a| a.kind_of?(Automation) && a.type == :chef_automation }
        return !chef_automations.empty?
      end

      # Determines if an automation is included in a list of acceptable automations. 
      # All automations are considered to be in the empty list.
      # Args:
      #   -+automation+ the automation to check
			#   -+filter+ list of automations on the node that are desired; empty implies all
      def satisfies_automation_filter (automation, filter = [])
        filter.empty? || filter.include?(automation)
      end # def satisfies_automation_filter
      
      ############################
      #
      # THIS CODE BELOW IS ALL LEGACY AND ONLY USED IN RC2Deployer anymore where we still allowed 
      # to include references to Chef recipes. It should go away once we refactor RC2Deployers.
      #
      ############################
      
      # Generates the CHEF_NODE_ATTR value for a given node that defines what parameters
      # need to be overridden when invoking chef-solo on the target machine.
      # * *Args*    :
      #   -+node+ The node (infrastructure of application) defining that has Chef scripts in
      #           on if their components.
      # * *Returns* :
      #   -+String+ A JSON formatted string containing all the overridden attributes.
      #             It needs to be added to the node.json file when invoking chef-solo.
      def generate_chef_node_attr(node)
        node_attr = {}
        collect_attributes(node, node_attr)
        result = ""
        i = 0
        length = node_attr.size()
        node_attr.each do |k,v|
          result << "\"#{k}\":#{v.to_json}"
          result << "," if i < length - 1 # only append if not the last
          i += 1
        end
        return result
      end

      # Recursively collects all chef attributes that need to be put on the CHEF_NODE_ATTR file
      # from nodes (and implicitly their links).
      # * *Args*    :
      #   -+node+ The node (infrastructure of application) defining that has Chef scripts in
      #           on if their components.
      #   -+node_attrs+ The node attributes that have been collected already
      # * *Returns* :
      #   -+String+ A Hash of attributes for CHEF
      def collect_attributes(node, node_attrs = {})

        # check for node level attributes that do not have a value
        node.amd.each do |k,v|
          if v.mutable && v.required && v.value && !v.implicit? == nil
            raise WeaverError, "Attribute '#{k}' does not have a value assigned."
          end
        end

        node.components.each do |c_id, c|
          # check for component level attributes that do not have a value
          c.amd.each do |k,v|
            if v.mutable && v.required && v.value == nil
              raise WeaverError, "Attribute #{v.fully_qualified_id} does not have a value assigned."
            end
          end

          c.automation.each do |a_id, s|
            # this should go into a script handler
            if s.script_type == :chef
              @@logger.debug "Processing script: #{s.executable}"
              s.amd.each do |k, v|
#                val = v.value.kind_of?(LateBinding) ? v.value.weaver_eval(c) : v.value
                val = v.value
                @@logger.debug("Processing chef parameter: #{v.name} = #{val}")

                # split up the attribute string for chef
                split = v.name.to_s.split(".")  
                
                #"nagios.new_client.client_address"
                map = node_attrs
                split.each_with_index do |elem, i|
                  if map[elem] == nil 
                    if i < split.size-1 # if we are not at the second last add Hash
                      map[elem] = Hash.new
                    else
                      map[elem] = val # add last element as value
                    end
                    map = map[elem]
                  else
                    map = map[elem]
                  end
                end                
              end
            end
          end
          collect_attributes(c, node_attrs)
        end
        return node_attrs
      end
      
      ######
      # END LEGACY CODE
      #######
      
      # Given a Hash of chef node attributes, return a String of the attributes in the format used by the json request. 
      # If a type other than Hash is passed in, it will be returned unmodified.
      #
      # * *Args*    :
      #   - +chef_node_attr+ -> a Hash of chef node attributes (or any other type, that will be returned unmodified)
      # * *Returns* :
      #   - A String of the attributes in the json format (for a Hash), or the chef_node_attr that was passed in (for any other type)
      #
      def self.convert_chef_node_attr_to_s(chef_node_attr)
        if (chef_node_attr.nil?)
          return nil
        elsif chef_node_attr.kind_of?(Hash)
          result = ""
          i = 0
          length = chef_node_attr.size
          chef_node_attr.each do |k,v|
            result << "\"#{k}\":#{v.to_json}"
            result << "," if i < length - 1 # only append if not the last
            i += 1
          end
          return result == "" ? " " : result # the " " is important otherwise IWD thinks that we did not set the value but we want an empty chef_node_attr
        elsif chef_node_attr.kind_of?(String)
          return chef_node_attr == "" ? " " : chef_node_attr # the " " is important otherwise IWD thinks that we did not set the value but we want an empty chef_node_attr
        else
          return " "
        end
      end
      
      def self.interpolate_and_convert_chef_node_attr_to_s(chef_node_attr, sandbox)
        if chef_node_attr.kind_of? Hash
          interpolate_properties(chef_node_attr, sandbox)
          return convert_chef_node_attr_to_s(chef_node_attr)
        end
        return chef_node_attr
      end
      
      def self.escape_quotes(str)
        return str.gsub(/\"/, %Q{\\"})
      end # def self.escape_quotes
      
      def self.interpolate_properties(properties, sandbox)
        unless properties.nil?
          properties.each do | k, v |
            unless v.nil?
              if (v.kind_of?(String) && !sandbox.nil?)
                regexp = /\#{([^\}]*)}/
                matchdata = regexp.match(v.strip)
                unless matchdata.nil?
                  interpolate = matchdata[1]
                  @@logger.debug "About to interpolate property #{interpolate}"
                  # sandbox.instance_exec() { @dependencies.clear } # clear dependencies to avoid a recursive reference error
                  entity = sandbox.instance_exec() { @entity }    # get member variable
              properties[k] = Sandbox.new(entity, nil, '', {}, {}, AllCache.new).instance_exec(v) { |str| eval("\"#{str}\"") }
                  @@logger.debug "Interpolated  property is #{properties[k]}"
                end
              end
              if v.kind_of? Hash
                interpolate_properties(v,sandbox)
              end
            end
          end
        end
      end

    end
  end
end