require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Component which may contain Automation
    # 
    # Author:: Ed Snible
    module AutomationContainer
      # Let all classes that include AutomationContainer have a method to return their Automations
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Automation
          def automation
            # Since mixins have no initialize() to set up @automation, initialize @automation lazily
            @automation ||= {} unless frozen?
            return @automation || {}
          end
          
          # Automation order is the order in which scripts are executed when the VM implementing the node is booted
          # Array whose members are Automation or arrays of Automation
          def automation_order
            @automation_order ||= [] unless frozen?
            return @automation_order || []
          end
          
          # Automation order is the order in which scripts are executed when the VM implementing the node is booted
          # * +automation_order+ Array whose members are Automation or arrays of Automation
          def automation_order=(automation_order)
            @automation_order = automation_order
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_automation)
        baseclazz.register_validate_handler(:validate_automation)
        baseclazz.register_child_enumerator(:automation)
        baseclazz.register_reference_freshener(:freshen_automation_order)           
      end
      
      # Deeply copy @automation
      def initialize_copy_automation(other_automation_container)
        deep_automation_copies = {}
          
        # Deeply copy @automation
        @automation = {}
        other_automation_container.automation.each { |id, automation|
          dup = automation.clone()           # As a side effect this will zap the looked-up member IDs
          dup.scope = self
          @automation[id] = dup
          deep_automation_copies[automation] = dup
        }

        # We will make the deep copy of @automation_order later during freshen_automation_order
      end
      
      # Validate @automation
      def validate_automation(out_messages, desired_phases = nil)
        automation.each { |id, automation| automation.validate_collect(out_messages, desired_phases) }
      end
      
      # Called as part of Duplicate.duplicate to ensure Topology.startup_order refers to the duplicate, not original, nodes      
      def freshen_automation_order(old_to_new)
        unless @automation_order.nil?
          @automation_order = @automation_order.map { |ele| ele.respond_to?(:map) ? ele.map { |node| old_to_new.key?(node) ? old_to_new[node] : node } : old_to_new.key?(ele) ? old_to_new[ele] : ele }
        end
      end
    end
    
  end
end