require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Application which may contain Group
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module GroupContainerMixin
      # Let all classes that include GroupContainerMixin have a method to return their groups
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Group
          def groups
            # Since mixins have no initialize() to set up @groups, initialize @groups lazily
            @groups ||= {} unless frozen?
            return @groups || {}
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_groups)
        baseclazz.register_validate_handler(:validate_groups)
        baseclazz.register_child_enumerator(:groups)           
        baseclazz.register_reference_freshener(:member_freshener)
        
      end
      
      # Deeply copy @groups
      def initialize_copy_groups(other_group_container)
        # At this point, self.initialize_copy should have cloned @groups but the values are shallow.
        @groups = {}
        other_group_container.groups.each do |id, group|
          dup = group.clone()
          dup.scope = self
          @groups[id] = dup
        end
      end
      
      # Validates contained #groups
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_groups(out_messages, desired_phases = nil)
        groups.each { |id, group| group.validate_collect(out_messages, desired_phases) }
      end
      
     # Finish the work started in duplicate and fix up membership references
      def member_freshener(old_to_new)
        groups.each do |group_id, group|
          group.members.each do |member_id, member|
            old_group = old_to_new.select { |old,new| old.is_a?(Weaver::Core::Group) and group_id  == old.id }.keys.first
            unless old_group.nil?
              old_member = old_group.members.select { |omid, omember| omid == member_id }.values.first
              new_member = old_to_new[old_member]
              group.members[member_id] = new_member
            end
          end # group.members.each do
        end # groups.each do
      end # def membership_freshener
      
      # Return a String containing a Weaver description of the group with members
      def to_weaver_groups
        # derive template name according to convention
        template_file = 'data/templates/weaver_core_groupcontainer.erb'
        file = File.join(File.dirname(__FILE__), "..", "..", "..", template_file)
        unless File.exists?(file)
          raise WeaverError, "Cannot find Erubis template file #{template_file} for mixin GroupContainerMixin."
        end
        return Weaver::Utils::Indent.fixup(Erubis::Eruby.new(File.read(file)).result(binding))
      end

    end
    
  end
end