require 'weaver/core/base'
require 'weaver/core/component_container_mixin'
require 'weaver/core/policy_container_mixin'
require 'weaver/core/provider_container_mixin'
require 'weaver/core/automation'
require 'weaver/core/automation_container'
require 'weaver/core/depends_on_mixin'
require 'weaver/core/operating_system'
require 'weaver/core/network_interface_controller'
require 'weaver/core/service_container_mixin'
require 'resolv'

module Weaver
  module Core
    # Represents a physical or virtual machine with an Operating System
    # @see Weaver::Dsl::NodeStatement
    #
    # Author:: Ed Snible
    # Author:: Florian Rosenberg
    class Node < Base
      include Weaver::Core::Generation
      include ServiceContainerMixin
      include ProviderContainerMixin
      include PolicyContainerMixin
      include AutomationContainer
      include ComponentContainerMixin
      include DependsOnMixin

      # The NetworkInterfaceControllers are children of Nodes and should appear in #flatten_children calls
      register_child_enumerator :nics
      
      # NICS should be deeply copied 
      register_initialize_copy_handler :initialize_copy_nics
      
      # NICS should be validated
      register_validate_handler :validate_nics
      
      # After cloning Node.realized_by is stale, this method updates it
      register_reference_freshener :freshen_realized_by
      
      # A :nic0 to use when unattached to a Root.  (If we are attached to a Root, use Root's :nic Weaver extended type)
      MasterNic0 = NetworkInterfaceController.new(:nic0, nil)

      # @return [Cloud] Data about the target cloud for this node
      attr_accessor :cloud

      # @return [Node[]]
      # The Node(s) that realize this Node.  Typically
      # application nodes are realized on physical environment nodes.
      # +[]+ if this node is not realized
      attr_reader :realized_by

      # @return [OperatingSystem]
      # The 'real' os if realized_by.nil?, otherwise
      # the OS requirement of this node.
      attr_accessor :os

      # @return [Firewall] The firewall ports that may pass
      attr_accessor :firewall

      # @return [Requirement[]]
      # Array of requirements for deploying Node successfully
      attr_accessor :requirements
      
      # @return [Hash]
      # Hash of id -> NetworkInterfaceController
      attr_accessor :nics
      
      # Used to hold raw responses from deployment system
      attr_accessor :cloud_status_internal
      
      # * +id+ A symbol for the Node, e.g. +:node1+
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        super(id, parent)
        
        @type = :node
        @requirements = []
        @realized_by = []
        @os = OperatingSystem.new
        @os.type = :unknown
        @firewall = nil
        @provisioner = nil
        @cloud_status_internal = {}
        
        # NICs should have a provenance so their attributes appear in short form, make sure we have one
        # whether or not we have a full Weaver tree with Root
        if parent.nil? || parent.root_entity.nil?
          @nics = { :nic0 => MasterNic0.clone }
          @nics[:nic0].provenance = MasterNic0
          @nics[:nic0].scope = self
        else
          @nics = { :nic0 => parent.root_entity.prototype_instance(:nic, :nic0, self, nil) }
        end

        # For hostname we have BOTH and attribute and Node-specific methods
        amd[:hostname] = Attribute.new(:hostname, self)
        amd[:hostname].value = LateBinding.new('nic0.hostname', self) { nic0.hostname }
        amd[:hostname].implicit = true
        amd[:hostname].required = false
          
        # For ip_address we have BOTH and attribute and Node-specific methods
        amd[:ip_address] = Attribute.new(:ip_address, self)
        amd[:ip_address].value = LateBinding.new('nic0.ip_address', self) { nic0.ip_address }
        amd[:ip_address].implicit = true
        amd[:ip_address].required = false
          
        attr_multiplicity = Attribute.new(:multiplicity, self, true)
        attr_multiplicity.value = 1
        attr_multiplicity.add_validator(:value_type, :numeric)
        attr_multiplicity.implicit = true
        attr_multiplicity.required = false
        amd[:multiplicity] = attr_multiplicity 
          
        amd[:multiplicity_index] = Attribute.new(:multiplicity_index, self)
        amd[:multiplicity_index].value = 0
        amd[:multiplicity_index].implicit = true
        amd[:multiplicity_index].required = false
      end

      # Make sure Node#clone creates a deep copy not a shallow copy
      # * +other+ Another Node
      def initialize_copy(other)
        super

        @requirements = []
        other.requirements.each do |requirement|
          dup = requirement.clone()
          dup.scope = self
          @requirements.push(dup)
        end

        initialize_copy_policies(other)

        os = other.os.clone()

        @realized_by = other.realized_by.clone
        # At this point, realized_by is a deep copy but refers to non-deep references
      end
      
      # * +other+ a Node
      def initialize_copy_nics(other)
        # At this point, self.initialize_copy should have cloned @nics but the values are shallow.
        @nics = {}
        other.nics.each do |id, nic|
          dup = nic.clone()
          dup.scope = self
          @nics[id] = dup
        end
      end

      # Update @realized_by after :initialize_copy has shallowly copied it 
      # * +old_to_new+ Hash of original Base entities to their deep-copied replacements
      def freshen_realized_by(old_to_new)
        @realized_by.map! { |oldtarget| old_to_new.key?(oldtarget) ? old_to_new[oldtarget] : oldtarget }
      end

      def validate_collect(out_messages, desired_phases = nil)
        @@logger.debug "Validating node '#{id}'"
        @@logger.debug "  has type '#{type}'"
        @@logger.debug "  has name '#{name}'"
        @@logger.debug "  has #{policies.size} polic(y/ies)"

        super

        # Currently these are validated in isolation.
        # TODO validate correctly, so if two Nodes with requirement for port 8080 are realized
        # upon the same infrastructure a message is generated here
        requirements.each do |requirement|
          requirement.validate_collect(out_messages, desired_phases)
        end
        
        validate_policies(out_messages, desired_phases);

        os.validate(out_messages, desired_phases)
      end

      # Returns true if this Node is equal to another Node
      def ==(other)
        retval = super
        return retval && @cloud == other.cloud && @components == other.components &&
        @requirements == other.requirements &&
        @realized_by == other.realized_by && @os == other.os
      end

      # Validate the contained NetworkInterfaceController s
      def validate_nics(out_messages, desired_phases = nil)
        nics.each { |id, nic| nic.validate_collect(out_messages, desired_phases) }
      end
      
      # Set hostname for node.  If node has more than one hostname, see Node#nics
      #  * +new_hostname+: the hostname (String) of this node.
      def hostname=(new_hostname)
        nics[:nic0].hostnames.clear
        nics[:nic0].hostnames << new_hostname
          
        if new_hostname != nil && nics[:nic0].ip_address.nil?
          begin
            nics[:nic0].ip_address = Resolv.getaddress(new_hostname)
          rescue
            nics[:nic0].add_validation_failure("Unknown host #{new_hostname}")
          end
        end
      end
      
      # Set IP Address for node.  See Node#nics
      #  * +new_ip_address+: the IP Address (String) of this node.
      def ip_address=(new_ip_address)
        nics[:nic0].ip_address = new_ip_address
      end
      
      # Return the controlling Provider.  If there are multiple ones
      # the provisioner is ambiguous but we will return one anyway, leaving any detection
      # of the problem to validation.
      def provisioner
        return @provisioner unless @provisioner.nil?
        return providers.values.to_a[0] unless providers.empty?  
#        return scope.provisioner
        return nil
      end
      
      # Set Provisioner / Cloud Provider for node
      # * +new_prov+ the Provisioner
      def provisioner=(new_prov)
        @provisioner = new_prov
      end
      
    end
  end
end