require 'weaver/core/logging'
require 'weaver/core/late_binding'
require 'weaver/core/validator'
require 'weaver/core/base'
require 'weaver/core/attribute'
require 'weaver/core/encrypted_value'
require 'weaver/dsl/unknown_statement_error'
require 'weaver/dsl/rule_statement'
require 'set'
require 'yaml'


module Weaver
  module Dsl
    include Weaver::Core

    # A base for the other statement classes.  Includes processing for attributes and the keywords
    # associated with symbols.
    #
    # Author:: Ed Snible
    # Author:: Florian Rosenberg
    class AbstractStatement
      include Weaver::Core
      
      # A class variable holding statement handler for handling dynamic types.
      @@statement_handlers = []           

      # Model represents the Core::Base that corresponds to the statement
      attr_accessor :model
      
      # 'parent' represents the enclosing \AbstractStatement
      attr_reader :parent
      
      # The WeaverParser evaluating this statement
      attr_reader :parser
      
      # * +id+ Symbol representing the entity, or a Core::Base which is the entity
      # * +parent_statement+ The enclosing AbstractStatement
      # * +weaver_parser+ The WeaverParser
      def initialize(id_or_model, type, parent_statement, weaver_parser = nil)
        raise "IllegalArgumentException" unless parent_statement.nil? || 
                                                parent_statement.kind_of?(AbstractStatement)

        @parser = weaver_parser.nil? ? parent_statement.parser : weaver_parser
        if id_or_model.kind_of?(Weaver::Core::Base)
          @model = id_or_model
          created_deeply = false
        else
          created_deeply = id_or_model.kind_of?(Symbol) && parent_statement.model.child_entities[id_or_model].nil?
          @model = parent_statement.model.lookup_or_define(id_or_model, type, parent_statement.weaver_stackframe) unless parent_statement.nil?
        end
        @parent = parent_statement
        
        # record newly created model
        unless id_or_model.kind_of?(Weaver::Core::Base)
         # @@logger.info "Multiple declarations for #{id_or_model}; '#{weaver_stackframe}' redefines entity originally defined at '#{@parser.definitions[id_or_model].weaver_origin}'" if @parser.definitions.key?(id_or_model)
          @parser.definitions[id_or_model] = @model
        end
        
        @expression_context = false
      end

      # The directory of the file owning this statement.  RECURSIVE      
      def dir
        return @dir unless @dir.nil?  # This file was directly parsed
        return nil if parent.nil?     # failure condition, no parsed root node
        return parent.dir             # recursively call parent statement's dir()
      end
      
      # Set the working directory for this statement.  Includes and imports with relative pathnames will be evaluated against +newdir+. 
      def dir=(newdir)
        @dir = newdir
      end

      # :stopdoc:
      # Process keyword +name+.  Name sets the name of an entity, or returns the name of the entity
      # * *Args* :
      #   - +name+ -> (optional) new name.  No action is taken if this argument is skipped
      def name(name = :rvalue_context)
        unless name == :rvalue_context
          @@logger.debug "name[#{name}]"
          model.name = name
        end 
        return model.name
      end

      # Process keyword +description+.
      # * *Args* :
      #   - +description+ -> (optional) new description.  No action is taken if this argument is skipped
      def description(description = :rvalue_context)
        unless description == :rvalue_context
          @@logger.debug "description[#{description}]"
          model.description = description
        end 
        return model.description
      end
      # :startdoc:

      # Process keyword +property+ to define or augment a Core::Attribute.  Properties are by default immutable.
      # The +attributedata+ hash may have keys
      # * +:name+ name used for automation, for example scripting or IWD names
      # * +:group+ group name (used for automation for IWD)
      # * +:validators+ array of expressions that evaluate to invocations of {validators}[rdoc-ref:Core::Validator]
      # * +:required+ if +true+, it shall be an error to deploy if the property is +nil+
      # * +:description+ human-readable label
      # * +:default+ default value if no actual value is supplied
      #
      # * *Args* :
      #   - +id+ -> Symbol naming the property
      #   - +attributedata+ -> Optional Hash with metadata about the property.
      #   - +block+ optional additional block with further metadata about the property
      def property(id, attributedata = {}, &block)
        attribute(id, attributedata, &block)
        
        if id.kind_of?(Hash)
          key = id.keys[0]
        else
          key = id
        end
        if key.kind_of?(String)
          key = key.symbolize
        end
        
        # TODO Warn if mutable was *explicitly* set to false
        attr = @model.amd[key]
        attr.mutable = true
      end
      
      # Process keyword +attribute+ which is an immutable property.
      #
      # @see #property
      #
      # See {+attribute+}[link:STATEMENTS_rdoc.html#label-attribute]
      # * *Args* :
      #   - +id+ -> Symbol naming the attribute
      #   - +attributedata+ -> Optional Hash with metadata about the property.
      #   - +block+ optional additional block with further metadata about the property
      def attribute(id, attributedata = {}, &block)
        if id.kind_of?(Hash)
          orig_hashed_id = id   # Fix for JRuby-1.6.5 optimizer problem, explicit copy of original value needed later
          key = id.keys[0]
          
          # if we have a default value for the attribute specified with => we
          # get the medatadata in the first parameter 'id' Hash, rather than
          # in the second parameter.  Copy them into attributedata
          id.each { |id, val| attributedata[id] = val unless id == key }
          attributedata[:value] = orig_hashed_id.values[0]
          id = key
        end

        # Make "attribute 'foo'" a shorthand for "attribute :foo, name=>'foo'"      
        if id.kind_of?(String)
          attributedata[:name] = id
          id = id.symbolize
        end
        
        @@logger.debug "attribute[id: #{id} on #{model.type} :#{model.id}]"
        
        value_override_blocked = !@model.amd[id].nil? && !@model.amd[id].mutable && !@model.amd[id].implicit?
        attr = @model.lookup_or_define(id, Attribute, weaver_stackframe)
        @model.amd[id] = attr
        
        attributedata.each { |key, value|
          if key == :value
            if value_override_blocked
              # We can't override an immutable value but should make a validation error
              @model.add_validator(:always_fail, "attributes[:#{id}] is not mutable and cannot be set to #{value}") unless attr.value == value
#            elsif value.kind_of?(String) && value.index('#{') != nil
#              code = 'Proc.new { ' + value.dump + ' }'
#              code = code.sub('\\#', '#')
#              attr.value = LateBinding.new(nil, model, &(eval(code)))
            else
              attr.value = value
            end
          elsif key == :validators
            if value.kind_of?(Array)
              value.each { |validator_or_valexpr| attach_validator_to_attribute(attr, validator_or_valexpr) unless validator_or_valexpr == :weaver_unknown_statement }
            else
              attach_validator_to_attribute(attr, value)
            end
          elsif key == :type
            @@logger.warn ":type syntax deprecated; #{attr.fully_qualified_id} should use validators[value_type(:#{value})]"
            attr.constrain_value_type(value)
          else
            writemethod = key.to_s + "="
            if attr.class.method_defined?(writemethod) then
              if key == :valid_values
                @@logger.warn ":valid_values syntax deprecated; use validators[valid_values(#{value})]"
              end
              method = attr.method(writemethod)
              method.call(value)
            else
              @@logger.warn "Unknown attribute property #{key.inspect} on #{attr.fully_qualified_id}"
              model.add_validation_failure("Unknown 'attribute' statement property: ':#{key}'.")
            end
          end
        }
        
        unless block.nil?
          attrib_statement = AttributeStatement.new(id, self)
          attrib_statement.instance_eval &block
        end
        
        if attr.default.nil?
          attr.default = attr.value unless attr.unresolved_value.kind_of?(LateBinding) 
        else
          if attr.unresolved_value.kind_of?(LateBinding)
            @@logger.info "#{attr.fully_qualified_id} default value #{attr.default.inspect} will be ignored because the attribute is bound to #{attr.unresolved_value}"
          elsif attr.value.nil?
            attr.value = attr.default
          end
        end
        
        attr.implicit = false
      end
      
      # Process keyword +validators+
      #
      # @see #validator
      #
      # See {+validators+}[link:STATEMENTS_rdoc.html#label-validators]
      # * *Args* :
      #   - +list+ -> Array of validation invocations
      def validators(list)
        raise WeaverError.new("#{model.type} (:#{model.id}) validators expects an array") unless list.kind_of?(Array)
        list.each { |ele| validator(ele) }
      end
      
      # Process keyword +validator+.  This keyword is used both to define new validator and their custom code
      # as well as to attach an instance of a validator to a Core::Base.
      #
      # When +validator+ is being used to define a validator, +id_and_type+ is a hash of the {validator's}[Core::Validator]
      # ID to the type of entity it can be attached to, such as <code>:equals => :attribute</code>.  See #define_validator
      # for that case.
      #
      # When +validator+ is being used to attach a parameterless validator instance, such as <code>:validators => :not_empty</code>,
      # +id_and_type+ contains the Symbol +:not_empty+.
      #
      # When +validator+ is being used to attach a parameterized validator instance, such as <code>:validators => [valid_values(nil, 6, 8, 9, 2, 5)]</code>,
      # we don't get any kind of expression tree, we get an array representing the validator instance in _prefix_ form: <code>[:valid_values, nil, 6, 8, 9, 2, 5]</code>.
      # (The code to convert from the regular \Weaver function notation to validator prefix form is in #method_missing.  The code must be there because the \Weaver DSL
      # needs help from Ruby's parser so that the validator invocation seems to have valid syntax.)
      #
      # @see #validators
      #
      # See {+validator+}[link:STATEMENTS_rdoc.html#label-validator]
      # * *Args* :
      #   - +id_and_type+ -> Symbol | <ARRAY> | Hash
      #   - +block+ -> Block of code used when defining a validator
      def validator(id_and_type=nil, &block)
        # Do we have something like "validator :empty_password"
        if id_and_type.kind_of?(Symbol)
          model.add_validator(id_and_type)
          return
        end
        
        # Do we have something like "validator [:contains, :application]"
        if id_and_type.kind_of?(Array) && id_and_type.length > 0 && id_and_type[0].kind_of?(Symbol)
          model.add_validator(id_and_type[0], *id_and_type.slice(1..-1))
          return
        end
        
        # We must have something like validator ":empty_password => :application do .. end"
        # or something like validator ":empty_password => :application, :phase => :predeployment do .. end"
        define_validator(id_and_type, &block)
      end
      
      # Create a new Core::Validator
      #
      # See {+validator+}[link:STATEMENTS_rdoc.html#label-validator]
      # * *Args* :
      #   - +id_and_type+ -> Hash of ID to Symbol, where the symbol is +:any+ or a \Weaver type such as +:attribute:
      #   - +block+ -> Block of code which receives the entity to validate as the first parameter and calls 
      #     <code>report()</code> if there is a problem.
      def define_validator(id_and_type, &block)
        return model.add_validation_failure("Expected :validator_name => :validator_type but got #{id_and_type}"
          ) unless id_and_type.kind_of?(Hash)
        return model.add_validation_failure("Expected :validator_name => :type but got #{id_and_type.class} #{id_and_type}"
          ) unless id_and_type.size <= 2 && id_and_type.size > 0
        return model.add_validation_failure("Reserved keyword ':phase' cannot be a validator name"
                  ) unless id_and_type.keys[0] != :phase

        validator_symbol = id_and_type.keys[0]
        validator_type = id_and_type.values[0]
        if (id_and_type.size == 2)
           return model.add_validation_failure("Expected :validator_name => :validator_type, :phase => :validator_phase but got #{id_and_type.inspect}"
            ) unless id_and_type.keys[1] == :phase
           validator_phase = id_and_type.values[1]
        else
           validator_phase = :any
        end
        
        @@logger.debug "define validator[#{validator_symbol} for #{validator_type} and phase #{validator_phase}]"
        
        return model.add_validation_failure("validator #{validator_symbol} defined without block") if block.nil? 
        
        return model.add_validation_failure("It is not permitted to redefine validators (ID: :#{validator_symbol})"
          ) unless root_statement.model.named_validators[validator_symbol].nil? 
            
        new_val = Validator.new(validator_symbol, validator_type, block)
        new_val.phase = validator_phase
        root_statement.model.named_validators[validator_symbol] = new_val
      end
      
      # Process keyword +rule+ which defines a new root-level Core::Rule.
      #
      # See {+rule+}[link:STATEMENTS_rdoc.html#label-rule]
      # * *Args* :
      #   - +id+ -> Symbol naming the rule
      #   - +block+ -> Block containing a Weaver declaration describing the rule and including a Proc
      def rule(id, &block)
        @@logger.debug "Processing rule with id '#{id}'"
        rule_statement = RuleStatement.new(id, root_statement)
        rule_statement.model.weaver_origin = weaver_stackframe
        root_statement.model.rules[id] = rule_statement.model
        rule_statement.instance_eval(&block)
        unless rule_statement.model.contval.nil?
          artifacts = rule_statement.model.contval.artifacts 
          unless artifacts.nil?
            artifacts.each do |a|
              if root_statement.model.artifacts[a].nil?
                model.add_validation_failure("Rule :#{id} makes an invalid artifact reference: :#{a}")
                return
              end
            end
          end
        end
        if rule_statement.model.where.nil?
           model.add_validation_failure("Rule :#{id} has no 'where' clause")
           return 
        end
        if rule_statement.model.body.nil? && rule_statement.model.contval.nil?
          if rule_statement.model.phase.nil?
            model.add_validation_failure("Rule :#{id} must have either a 'body' or a 'continuous_validation' block")
          else
            model.add_validation_failure("Rule :#{id} has no 'body'") if rule_statement.model.phase != :continuous
            model.add_validation_failure("Rule :#{id} has no 'continuous_validation' block") if rule_statement.model.phase == :continuous
          end 
        elsif !rule_statement.model.body.nil?
          # A rule with a body cannot be continuous
           model.add_validation_failure("Rule :#{id} has a body; therefore, its phase cannot be :continuous") if rule_statement.model.phase == :continuous
        elsif !rule_statement.model.contval.nil?
           if rule_statement.model.phase.nil?
             # Rules that contain a 'continuous_validation' block are :continuous implicitly; so let's enforce it.
             rule_statement.model.phase = :continuous
           else
             # If the user explicitly set the phase of a rule with a 'continuous_validation' block, it must be :continuous
             model.add_validation_failure("Rule :#{id} has a 'continuous_validation' block; therefore, its phase must be :continuous") if rule_statement.model.phase != :continuous
           end            
        end
      end
            
      # Process keyword +include+ to include another document.  Unlike ImportContainerStatement#import,
      # included content is immediately available without #use.
      #
      # See {+include+}[link:STATEMENTS_rdoc.html#label-include]
      # * *Args* :
      #   - +filename+ -> filename.  If relative it is relative to the current file.
      def include(filename)
        unless filename.kind_of?(String)
          @@logger.warn "include expected a filename String but got a #{filename.class}"
          model.add_validation_failure("include expected a filename String but got a #{filename.class}", weaver_stackframe)
          return
        end
        
        @@logger.info "include[#{filename}] relative to #{dir.inspect}"
        begin
          @parser.parse_using_load_path(filename, dir, self.clone) # Use a clone in case parsing changes dir
        rescue SystemCallError => e
          @@logger.warn "include failed #{e.to_s}"
          model.add_validation_failure("include failed #{e.to_s}", weaver_stackframe(e.backtrace))
        end
      end

      # Process keyword <code>member_of [list of group symbols]</code>
      #
      # @see GroupStatement
      #
      # See {+member_of+}[link:STATEMENTS_rdoc.html#label-member_of]
      # * *Args* :
      #   - +containers+ -> Array of Symbol
      def member_of(*containers)
        containers.each { |group| @model.member_of[group] = nil }
        # Later during validation we will look up the symbols we saved here
      end
      
      # The language +use+ statement, to bring out something from a namespace, e.g. 'use test2.jke1 => :jke1'
      # After processing this statement, @model will contain a new entity whose ID is the value of ent_id_hash
      #
      # See {+use+}[link:STATEMENTS_rdoc.html#label-use]
      # * *Args*    :
      #   - +ent_id_hash+ -> a Hash from a Core::Base | Symbol to an new ID Symbol.
      #   - +block+ -> Optional block with further augmentations for the refined entity
      def use(use_expr, &block)
        
        if use_expr.kind_of?(Symbol)
          @@logger.debug "use #{use_expr} on #{model.type} :#{model.id}"
          sym = id = use_expr
        elsif (use_expr.kind_of?(Hash))
          id = use_expr.values.to_a[0]
          raw_entity = use_expr.keys.to_a[0]
          if raw_entity.kind_of?(AbstractStatement)
            @@logger.debug "use #{raw_entity.model.fully_qualified_id} on #{model.type} :#{model.id}"
            sym = nil
            entity = Weaver::Transforms::Duplicate.duplicate(raw_entity.model)
            entity.provenance = raw_entity.model
            entity.set_origin(weaver_stackframe, true)
          elsif raw_entity.kind_of?(Symbol)
            @@logger.debug "use #{raw_entity} on #{model.type} :#{model.id}"
            sym = raw_entity
          elsif raw_entity.kind_of?(LateBinding)
            @@logger.warn "Unknown source #{raw_entity.instance_eval {@expr} }"
            model.add_validation_failure("Unknown source #{raw_entity.instance_eval {@expr} }", weaver_stackframe)
            return
          else
            @@logger.warn "Unexpected use source type #{raw_entity.class}"
            model.add_validation_failure("Unexpected use source type #{raw_entity.class}", weaver_stackframe) unless raw_entity.kind_of?(FailureEater)
            return
          end
        else
          # FailureEaters are reported elsewhere and shouldn't get an addition error here
          unless use_expr.kind_of?(FailureEater)
            @@logger.warn "Unexpected use type #{use_expr.class}"
            model.add_validation_failure("Unexpected use type #{use_expr.class}")
          end
          return
        end

        unless sym.nil?        
          entity = model.import_by_symbol(sym)
          if entity.nil?
            @@logger.warn "Unknown use symbol #{sym.inspect}"
            model.add_validation_failure("Unexpected use symbol #{sym.inspect}")
            return
          else
            entity.set_origin(weaver_stackframe, true)
          end
        end

        entity.id = id
        entity.scope = model
        
        # We need to store the imported entity in the right container, so find it and call it
        @@statement_handlers.each do |func|
          return true if respond_to?(func) && send(func, entity, entity.id, &block)
        end
        
        @@logger.warn "'#{model.type}' is not a container for '#{entity.type}' so :#{entity.id} cannot be refined here"
        model.add_validation_failure("'#{model.type}' is not a container for '#{entity.type}' so :#{entity.id} cannot be used here")
        return false

      end
      
      # The language +rename+ statement which renames a direct or indirect child, e.g. 'rename core_os_1 :was_server'
      #
      # See {+rename+}[link:STATEMENTS_rdoc.html#label-rename]
      # * *Args*    :
      #   - +ent_id_hash+ -> a Hash from a Core::Base | Symbol to a new ID Symbol.
      def rename(ent_id_hash = nil)
#        @@logger.debug "rename #{ent_id_hash.inspect}"
        unless ent_id_hash.kind_of?(Hash)
          model.add_validation_failure("Syntax: rename source => :newname")
          return
        end
        
        new_id = ent_id_hash.values.to_a[0]
        entity = ent_id_hash.keys.to_a[0]
        
        if entity.kind_of?(AbstractStatement)
          entity = entity.model
        elsif entity.kind_of?(Symbol)
          old_id = entity
          @@logger.debug "rename :#{old_id} to #{new_id} on #{model.type} :#{model.id}"
          entity = model.child_entities[entity]
          if entity.nil?
            @@logger.warn "Unknown rename source :#{old_id}"
            model.add_validation_failure("Unknown rename source #{old_id.inspect}")
            return
          end
        elsif entity.kind_of?(LateBinding)
          # Weaver converts Attributes into LateBinding during parsing.  (#method_missing() does
          # it).  Reverse the LateBinding back into an Attribute
          expr = entity.to_s
          # expr is usually something like '#{facade_693.was_server.password}'
          md = /#\{(.*)\.([^\.]+)\}/.match(expr)
          if md.nil?
            @@logger.warn "rename #{entity} to :#{new_id} cannot rename"
            model.add_validation_failure("Cannot rename #{expr}")
            return
          else
            scope_expr = md[1]
            attr_expr = md[2]
            scope_entity = model.instance_eval scope_expr
            entity = scope_entity.amd[attr_expr.symbolize]
          end
        end

        @@logger.debug "rename #{entity} to :#{new_id} on #{entity.scope.type} :#{entity.scope.id}"
        begin
          entity.scope.rename(entity, new_id)
        rescue WeaverError => we
          entity.add_validation_failure(we.message)
        end
      end
      
      # The language +alias_of+ statement which creates an _alias_ for an arbitrary Core::Base entity.
      #
      # See {+alias_of+}[link:STATEMENTS_rdoc.html#label-alias_of]
      # * *Args*    :
      #   - +ent_id_hash+ -> a Hash from a Core::Base | Symbol to a new ID Symbol.
      def alias_of(ent_id_hash)
        new_id = ent_id_hash.values.to_a[0]
        entity = ent_id_hash.keys.to_a[0]
        
        if entity.kind_of?(AbstractStatement)
          entity = entity.model
        elsif entity.kind_of?(Symbol)
          old_id = entity
          @@logger.debug "alias :#{old_id} to #{new_id} on #{model.type} :#{model.id}"
          entity = model.child_entities[entity]
          if entity.nil?
            model.add_validation_failure("Unknown alias :#{old_id}")
            return
          end
        end
      
        @@logger.debug "alias #{entity} to :#{new_id} on #{entity.scope.type} :#{entity.scope.id}"
        model.alias_of(entity, new_id)
      end

      # The language +tag+ statement which adds a Symbol to an arbitrary Core::Base entity.
      #
      # See {+tag+}[link:STATEMENTS_rdoc.html#label-tag]
      # * *Args*    :
      #   - +symbol+ -> any Symbol; +:novalidate+ is an example of a symbol with special meaning
      def tag(symbol)
        @@logger.debug "tag #{symbol.inspect}"
        unless symbol.kind_of?(Symbol)
          model.add_validation_failure("Syntax: tag :<SYMBOL>")
          return
        end
        
        model.tags.add(symbol)
      end
      
      # If the +parent+ can be followed to a Root, returns that \Root.
      def root_statement
        statement = self
        while !statement.nil? && !statement.kind_of?(RootStatement)
          statement = statement.parent
        end
        return statement
      end      

      # The top visible statement.  This is usually the root statement but within an import it
      # can be the import which blocks access to the root
      def top_statement
        statement = self
        while !statement.nil? && !statement.kind_of?(RootStatement) && !statement.kind_of?(ImportStatement)
          statement = statement.parent
        end
        return statement
      end      

      # Process \Weaver language modifier +late_binding+
      #
      # @see Core::LateBinding
      #
      # See See {+late_binding+}[link:STATEMENTS_rdoc.html#label-late_binding]
      # * +late_binding+ a Block of Ruby or \Weaver code.
      def late_binding(&block)
        return LateBinding.new(nil, model, &block)
      end
      
      # Process \Weaver language helper +load_properties+
      #
      # * +filename+ the filename to load, which is expected to be a .properties or .yml filename
      def load_properties(filename = nil)
        unless filename.kind_of?(String)
          @@logger.warn "load_properties expected a filename String but got a #{filename.class}"
          model.add_validation_failure("load_properties expected a filename String but got a #{filename.class}")
          return {}
        end
        
        return load_java_properties(filename) if (File.extname(filename) == ".properties")
        return yaml_load_file(filename) if (File.extname(filename) == ".yml" || File.extname(filename) == ".yaml")
        
        @@logger.warn "load_properties expected a filename ending in .properties, .yml, or .yaml but got #{filename.inspect}"
        model.add_validation_failure("load_properties expected a filename ending in .properties, .yml, or .yaml but got #{filename.inspect}")
        return {}
      end
      
      # Process built-in function +decrypt+(ciphertext)
      #
      # See {+decrypt+}[link:STATEMENTS_rdoc.html#label-decrypt]
      # * +ciphertext+ Base64 encoded ciphertext
      def decrypt(ciphertext)
        return EncryptedValue.new(ciphertext)
      end
      
      # Process built-in function +config+ which provides access to config.yml or similar configuration Hash
      def config
        return parser.config
      end
      
      # The statement is part of an expression, such as 'foo.bar', not part of a declaration
      # such as 'node (:foo) { ... }'
      def expression_context?
        return @expression_context
      end

      # The statement is part of an expression, such as 'foo.bar', not part of a declaration
      # such as 'node (:foo) { ... }'
      # * +is_context+ set to <code>true</code> by the parent statement if encountered evaluating an expression
      def expression_context=(is_context)
        @expression_context = is_context
      end

      # Return +true+ and create a validation error if two (differently-typed) children are defined with the same id
      def fail_id_redefinition(id, clazz)      
        if model.child_entities.key?(id) && !model.child_entities[id].kind_of?(clazz)
          model.add_validation_failure("ID :#{id} already used for #{model.child_entities[id].type}")
          return true
        end
        return false
      end
     
	    # Registers statement handlers that are automatically invoked when the parser 
	    # looks for a statement that is unknown (i.e., extensible types). The 
      # method_missing uses the registered handlers to find an appropriate method. 
	    # This concept is used to encapsulate statements that occur in multiple other
      # statements as mixins (modules ending with *Container) to avoid duplication.
      def self.register_statement_handler(elements)
        # TODO Use technique like Base to have per-class statement_handlers instead of one shared by all statements
        elements.each { |element| @@statement_handlers.push(element) unless @@statement_handlers.member?(element) }
      end
  
      # :stopdoc:
      def method_missing(sym, *args, &block)
        @@statement_handlers.each do |func|
          if self.respond_to?(func) 
            x = self.send func, sym, args[0], &block
            if (x)
              return x
            end
          end
        end
        
        # Treat "x y" as assignment "@amd[:x].value = y"
        attr = @model.amd[sym]
        if !attr.nil?
          if args.length == 1
            unless block.nil?
              @@logger.warn "Unexpected block {} following assignment of #{model.fully_qualified_id}.#{sym} = #{args[0].inspect}"
              model.add_validation_failure("Unexpected block {} following assignment of #{model.fully_qualified_id}.#{sym} = #{args[0].inspect}")
              return
            end
            
            # Give the Ruby model a chance to override an Attribute's setter (used for ip_address=)
            # Ruby 1.8.x uses Strings for public_methods, 1.9 uses Symbols
            if model.public_methods.include?("#{sym}=") || model.public_methods.include?("#{sym}=".to_sym) 
              model.send "#{sym}=".to_sym, args[0]
              return
            end
            
            unless attr.mutable || attr.implicit?
              @@logger.warn "Attribute #{attr.fully_qualified_id} cannot be changed.  Only mutable properties can be changed"
              attr.add_validation_failure("Attribute cannot be changed.  Only mutable properties can be changed")
              return
            end
            
            attr.value = args[0]
            attr.weaver_origin = weaver_stackframe
            return
          end
        end
        
        # Process "x= y" by forwarding to the model
        if sym.to_s[-1,1] == '='
          # We are being asked to set, do we have an attribute by that name?
          attr = @model.amd[sym.to_s.chop.to_sym]
          unless attr.nil?
            unless attr.mutable || attr.implicit?
              @@logger.warn "Attribute #{attr.fully_qualified_id} cannot be changed.  Only mutable properties can be changed"
              attr.add_validation_failure("Attribute cannot be changed.  Only mutable properties can be changed")
              return
            end
            
            # Forward to the model to do the actual work
            @model.send sym, *args, &block
            # But the statement sets the origin
            attr.weaver_origin = weaver_stackframe
            return
          end
        end
          
        # Treat "x" as a synonym for a statement about scoped_lookup(:x), so that we can say
        #   entity.attribute = 'hello'
        # or even
        #   entity.component :newcomp { ... }
        #
       
        entity = @model.scoped_lookup(sym)
        unless entity.nil?
          return create_late_binding(sym) if entity.kind_of?(Attribute)

          statement = root_statement.statement_for_entity(entity, self)           
          statement.instance_eval &block unless block.nil?
          return statement
        end
        
        # return an ImportStatement if we see the model element for an Import
        if model.respond_to?(:weaver_imports) && model.weaver_imports.key?(sym)
          return root_statement.statement_for_entity(model.weaver_imports[sym], self)
        end
        
        namespace = top_statement.model.weaver_imports[sym]
        unless namespace.nil?
          return root_statement.statement_for_entity(namespace, self)
        end
        
        # Transform range(1, 10) into [:range, 1, 10]
        validator_proc = root_statement.model.named_validators[sym]
        unless validator_proc.nil?
          return [sym].concat(args)
        end
        
        # Is 'sym' a usable object in the anonymous import namespace?
        candidate_importable = model.import_by_symbol(sym)
        unless candidate_importable.nil?
          return root_statement.statement_for_entity(candidate_importable, self)
        end

        @@logger.warn "Unknown statement '#{sym}' on #{model.type} #{model.fully_qualified_id} processing #{self.class}, child names are #{model.child_entities.keys.inspect}"
        # We used to attach the 'Unknown Statement' failure to @model itself, but the line number information was being lost because @model's
        # line gets reset to the last time it was augmented (or used in a path expression).  Instead we create a pseudo-attribute
        # with the same name as the unexpected symbol, and let its weaver_origin keep the line number of the actual problem.
        pseudo_attribute = Attribute.new(sym, model)
        pseudo_attribute.implicit = false
        pseudo_attribute.weaver_origin = weaver_stackframe
        model.amd[sym] = pseudo_attribute
        pseudo_attribute.add_validation_failure("Unknown statement '#{sym}' on #{model.type} #{model.fully_qualified_id}", weaver_stackframe)
        return FailureEater.new(sym)  # FailureEater can consume not only 'foo' but 'foo.bar.baz'
      end
      # :startdoc:

      # The filename of the .weaver file containing this statement
      def weaver_filename
        statement = self
        while !statement.nil? && !statement.kind_of?(RootStatement) && !statement.kind_of?(ImportStatement)
          statement = statement.parent
        end
        return statement.nil? ? nil : statement.model.id
      end
      
      # Extract the .weaver filename and line number by inspecting a call stack.
      # * *Args*    :
      #   - +bt+ -> (optional) Array representing a stack frame.  If not supplied, uses Kernel#caller
      def weaver_stackframe(bt = caller)
        # 'pattern' represents the filename, followed by a colon, typically 'app.weaver:', as it appears in a stackframe,
        # it can also be something like 'redeploy.weaver.1100759452.6374:'
#        pattern = "#{root_statement.model.id}:"
        pattern = "#{weaver_filename}:"
        line_in_backtrace = bt.find {|stackframe| stackframe.include? pattern }
          
        # If we don't find the file we expected look for any .weaver: file, e.g. builtins.weaver
        line_in_backtrace = bt.find {|stackframe| stackframe.include? ".weaver:"} if line_in_backtrace.nil?
          
        return nil if line_in_backtrace.nil?
        md = /(?<sourcefile>.*):(?<lineno>[0-9]*):in `(.*parse.*|\(eval\))'/.match(line_in_backtrace)
        return md.nil? ? line_in_backtrace : "#{md[1]}:#{md[2].to_i}"
      end
      
      protected
      
      # Create a Core::LateBinding for +sym+ which references an Core::Attribute
      # * +sym+ An attribute of the <code>@model</code> of this statement
      def create_late_binding(sym)
        # Build the path that lead us to this point
        qual = []
        statement = self
        until statement.nil? || !statement.expression_context?
          qual << statement.model.id.to_s
          statement = statement.parent
        end
        qual.reverse!
        
        qual << sym.to_s
        
        expr = qual.join('.')
        code = "Proc.new { #{expr} }"
        @@logger.debug "Creating late_binding using #{code}"
        return LateBinding.new(expr, statement.model, &(eval code))
      end
      
      private
      
      # Read a YAML file.  Creates a validation failure on I/O rather than throwing an exception.
      # * *Args*    :
      #   - +filename+ -> String typically ending in .properties
      # * *Returns* Hash
      def yaml_load_file(filename)
        begin
          return YAML.load_file(filename)
        rescue StandardError => se
          warn_and_fail_validation("load_properties failed for #{filename} #{se.to_s}", weaver_stackframe(se.backtrace))
          return {}
        end
      end

      # Read a Java-style property file.  Creates a validation failure on I/O rather than throwing an exception.
      # * *Args*    :
      #   - +filename+ -> String typically ending in .properties
      # * *Returns* Hash
      def load_java_properties(filename)
        # Don't check for file-not-found, we will catch I/O problems instead
        retval = {}
          
        begin
          File.open(filename, 'r') do |file|
            file.each_line do |line|
              line_data = line.scan(/^(?i)(?!#)(.+?)=(.+)?/)[0]
              if !line_data.nil? && line_data.any?
                retval[line_data[0]] = line_data[1]
              end
            end
          end
          
        rescue SystemCallError => se
          warn_and_fail_validation("load_properties failed for .properties #{se.to_s}", weaver_stackframe(se.backtrace))
        end
        
        return nest_like_yaml(retval).merge!(retval)
      end
      
      def warn_and_fail_validation(msg, origin)
        @@logger.warn msg
        model.add_validation_failure(msg, origin)
      end
      
      # Given a Hash whose keys are Strings containing periods (.), create a Hash-of-Hashes breaking on the periods
      def nest_like_yaml(h)
        retval = {}
          
        h.each do |key, value|
          map = retval 
          split = key.split('.')
          split.each_with_index do |elem, i| 
            if i == split.length-1 # if we hit the last one -> actual value
              unless map.key?(elem) 
                map[elem] = value
                # Ignore collisions -- if elem is already in this level of retval, throw it away
              end
            else
              map[elem] = {} unless map.key?(elem)
            end
            
            map = map[elem]
          end
        end
        
        return retval
      end
      
      def attach_validator_to_attribute(attr, validator_or_expression_array)
        if validator_or_expression_array.kind_of?(Array)
          attr.add_validator(validator_or_expression_array[0], *validator_or_expression_array.slice(1..-1))
        else
          attr.add_validator(validator_or_expression_array)
        end
      end
      
      # Remove kernel methods to get something like a Blank Slate pattern so that
      # Weaver users don't need to know about and avoid obscure Ruby ::Object methods.       
      [:test, :methods, :define_singleton_method, :freeze, :untrust, :extend, :protected_methods,
      :method, :tap, :singleton_class, :public_method, :untaint, :display, :singleton_methods,
      :taint, :instance_variables, :public_methods, :rand, :gets, :readlines, :exec, :select,
      :system, :lambda, :throw, :fail, :catch, :abort, :require_relative, :proc, :srand,
      :set_trace_func, :exit, :open, :spawn, :trace_var, :autoload, :trap, :format, :fork,
      :syscall, :load, :warn, :binding, :sleep].each do |meth|
        begin 
          undef_method meth
        rescue NameError
          # Do nothing, the list is for Ruby 1.9.2 and the actual implementation might not have
          # all the methods.  It isn't sufficient to check #respond_to? because Object doesn't
          # respond_to? :test even though it's there.
        end
      end
      
    end # class AbstractStatement
    
    # This class helps the Ruby parser recover from an error by 'eating' certain semantic errors in a Weaver file,
    # especially <unknown>.<expression>.  It is not designed to be used or extended.
    class FailureEater
      # @private
      def initialize(name)  # :nodoc:
        @name = name
      end
      
      # @private
      def id # :nodoc:
        @name
      end
      
      # +:missing_entity+
      def type
        :missing_entity
      end
      
      # @private
      def to_s  # :nodoc:
        return "unknown '#{@name}'"
      end
      
      # @private
      def method_missing(sym, *args, &block)  # :nodoc:
        return self
      end
    end

  end # module Dsl
end # module weaver