require 'zk'
require 'timeout'
require 'weaver/system'
require 'weaver/core/logging'
require 'weaver/core/topology'
require 'weaver/core/attribute'
require 'weaver/core/node'
require 'weaver/core/weaver_error'
require 'weaver/transforms/fold'
require 'weaver/transforms/duplicate'
require 'weaver/transforms/unroll'
require 'weaver/transforms/merge'
require 'weaver/dsl/weaver_parser'
require 'weaver/deployers/deployer'
require 'weaver/providers/delegating_provider'
require 'weaver/utils/plugin_loader'
require 'weaver/deployers/execution/coordination_util'
require 'weaver_coordination'
require 'weaver/coordination/system'
require 'weaver/execution/change_plan'
require 'weaver/utils/project_helper'

module Weaver
  
  # @private
  module ConfigStatus
  
    def self.find_by_id id
      if id.instance_of? String
        id = id.to_i
      end 
      values.each do |type|
        if id == type.id
          return type
        end
      end
      nil
    end
  
    def self.find_by_value value
      values.each do |type|
        if value.eql?(type.label)
           return type
        end
      end
      nil
    end
  
    def self.values
      [UNKNOWN, CONFIGURING, CONFIGURED, DELETING, DELETED, FAILED] 
    end
  
    class Enum
      include Comparable
      attr_reader :id, :label
  
      def initialize id, label
        @id = id
        @label = label
      end
  
      def <=>(obj)
        return -1 if obj.nil?
        return -1 if @id < obj.id
        return 1 if @id > obj.id
        return 0
      end
    end # class Enum
  
    UNKNOWN = Enum.new(1, 'unknown')
    CONFIGURING = Enum.new(2, 'configuring')
    CONFIGURED = Enum.new(3, 'configured')
    DELETING = Enum.new(4, 'deleting')
    DELETED = Enum.new(5, 'deleted')
    FAILED = Enum.new(6, 'failed')
  
  end # module ConfigStatus
  
  # Environment is a facade into the Weaver system that can
  # parse an environment topology and deploy the system via
  # the appropriate deployer plugin.
  #
  # Author:: Florian Rosenberg
  # Author:: Ed Snible
  class Environment
    include Weaver::Core
    
    # load all plugins in the system.  We do this once, while parsing Environment, so that
    # we don't reload if we are doing many, many deployments
    Weaver::Utils::PluginLoader.new(
      [File.join("weaver", "plugins", "deployers")],
      [File.expand_path(File.join("deployers", "plugins"), File.dirname(__FILE__))]
    ).load_plugins
    
      # @return [Root] The Weaver model
      attr_accessor :root
      
      # @return [ProjectHelper] The project configuration
      attr_reader :project
      
    # Creates a new Environment from a given Weaver file.
    #
    # @param [Root|String] root_or_file the Weaver files containing the environment topology, or a parsed version of it
    # @param [String|Weaver::Utils::ProjectHelper|nil] config_filename filename of config.yml, or a loaded project
    # @param [Hash<String, Object>] config_overrides where keys in dotted notation refer to nested Hashes
    # @param reporter Optional callback
    #
    # @return the newly created Environment instance.
    #
    def initialize(root_or_file, config_filename = nil, config_overrides = {}, &reporter)

      if root_or_file.kind_of?(Root)
        raw = root_or_file
        
        if config_filename.nil?
          @project = Weaver::Utils::ProjectHelper.new(nil, config_overrides)
        elsif config_filename.kind_of?(Weaver::Utils::ProjectHelper)
          @project = config_filename
        else
          @project = Weaver::Utils::ProjectHelper.new(File.dirname(config_filename), config_overrides, config_filename)
        end
          
      else
        environment_topology_file = root_or_file
        unless File.exists?(environment_topology_file)        
          raise Weaver::Core::WeaverError, "Environment file does not exist: #{environment_topology_file}"   
        end
        
        if config_filename.nil?
          # if a .weaver file is nearby use that for project directory
          if File.exists?(File.join(File.dirname(environment_topology_file), '..', Weaver::Utils::ProjectHelper::WEAVER_FILE))
             @project = Weaver::Utils::ProjectHelper.new(File.join(File.dirname(environment_topology_file), '..'), config_overrides)
          elsif File.exists?(File.join(File.dirname(environment_topology_file), '..', '..', Weaver::Utils::ProjectHelper::WEAVER_FILE))
            @project = Weaver::Utils::ProjectHelper.new(File.join(File.dirname(environment_topology_file), '..', '..'), config_overrides)
          else
            # Environment can handle .weaver files without associated project and config, but they will lack config[] parameters
            @project = Weaver::Utils::ProjectHelper.new(nil, config_overrides)
          end
        elsif config_filename.kind_of?(Weaver::Utils::ProjectHelper)
          @project = config_filename
        else
          @project = Weaver::Utils::ProjectHelper.new(File.dirname(config_filename), config_overrides, config_filename)
        end
        
        raw = Weaver::Dsl::WeaverParser.new(@project.config_params).parse(environment_topology_file)
      end
      
      raw.environment = self
      
      # Create pseudo-validation-failures for config.yml overrides that didn't work
      @project.validation_messages.each { |msg| raw.add_validation_failure(msg) }
        
      @reporter = reporter || Proc.new { |lvl, status, msg| }
      begin
        folded = Weaver::Transforms::Fold.fold(raw)
        @root = Weaver::Transforms::Unroll.unroll(folded)
      rescue WeaverError => we
        @root = raw
        @root.add_validation_failure(we.message)
      end
      
      @instance_dir = "tmp/my_instances"
    end
    
    # Deeply copy on #clone
    def initialize_copy(other)
      super
      
      @root = Weaver::Transforms::Duplicate.duplicate(@root)
    end

    # Validates an environment
    #
    # * *Args*    :
    #   - +phases+ -> A specific list of validation phases (Array)
    #   - +properties+ -> a list of properties that are needed to configure the underlying deployer.
    #                     This is implementation specific and may contain username, password, key files, etc.
    #   - +deploy_parameters+ -> deployment time parameters for the environment topology (e.g., :deps_file_url)
    #
    # * *Returns* :
    #   A list of ValidationMessage instances containing possible validation problems
    #
    def validate(phases = [], properties = {}, deploy_parameters = {})
      e2 = simulate_deploy(properties, deploy_parameters)
               
      messages = []
      e2.root.validate_collect(messages, phases)
      return messages
    end
    
    # Validates an environment and produces a validation report a text file.
    #
    # * *Args*    :
    #   - +filename+ -> filename where to store the report
    #   - +phases+ -> A specific list of validation phases (Array)
    #   - +properties+ -> a list of properties that are needed to configure the underlying deployer.
    #                     This is implementation specific and may contain username, password, key files, etc.
    #   - +deploy_parameters+ -> deployment time parameters for the environment topology (e.g., :deps_file_url)
    #
    # * *Returns* :
    #   Two numbers, the first one expressing the number of error, the second the number of warnings. 
    #
    def validate_report(filename, phases = [], properties = {}, deploy_parameters = {}) 
      messages = validate(phases, properties, deploy_parameters)
      error_count = warn_count = 0
      File.open(filename, "w") do |f|
        messages.each do |msg| 
          msg_str = msg.to_s
          f.write "#{msg_str}\n"
          if / ERROR / =~ msg_str || / FATAL / =~ msg_str
            error_count += 1
          elsif / WARNING / =~ msg_str
            warn_count += 1
          end
        end
      end 
      return error_count, warn_count
    end

    # Provisions the environment.
    #
    # * *Args*    :
    #   - +properties+ -> a list of properties that are needed to configure the underlying deployer.
    #                     This is implementation specific and may contain username, password, key files, etc.
    #   - +deploy_parameters+ -> deployment time parameters for the environment topology (e.g., :deps_file_url)
    #
    # * *Returns* : the update environment instance (type Environment)
    #
    # * *Raises* :
    #   - +WeaverError+ if an error occurs while deploying the environment
    #
    # TODO: remove this method when use templates
    def deploy(properties = {}, deploy_parameters = {})
      return update(nil, properties, deploy_parameters)
    end
    
    # Updates a deployed instance (or creates if not deployed)
    #
    # * *Args*    :
    #   - +instance_id+ -> deployed instance id if exists
    #   - +properties+ -> a list of properties that are needed to configure the underlying deployer.
    #                     This is implementation specific and may contain username, password, key files, etc.
    #   - +deploy_parameters+ -> deployment time parameters for the environment topology (e.g., :deps_file_url)
    #
    # * *Returns* : the update environment instance (type Environment)
    #
    # * *Raises* :
    #   - +WeaverError+ if an error occurs while deploying the environment
    #
    # TODO: remove this method when use_templates
    def update(instance_id, properties = {}, deploy_parameters = {})
      if topology.nil?
        raise WeaverError.new("No topology to deploy")
      end

      deployed_env = nil
      deployed_model = instance_id.nil? ? nil : "#{@instance_dir}/#{instance_id}/model.weaver"
      if !instance_id.nil?
        raise WeaverError.new("Deployed instance model cannot be found") unless File.exists?(deployed_model)
        deployed_env = Weaver::Dsl::WeaverParser.new.parse(deployed_model)
        raise WeaverError.new("Deployed instance model invalid") if deployed_env.nil?
        @root = Weaver::Transforms::Merge.merge(@root, deployed_env)
      end
            
      if deploy_parameters[:mock] == "true"
        deploy_parameters[:mock_coordination] = "true" if deploy_parameters[:mock_coordination].nil?
      end
      
      coordinator_configuration = nil
      transaction = nil
      timed_out = nil

      begin
        # TODO this can be eliminated if we eliminate the --timeout option
        project.config_params.merge!({:timeout => properties[:timeout]}) unless properties[:timeout].nil?
        @@logger.debug "timeout set to #{config[:timeout]}"
        Timeout::timeout(config[:timeout]) {
          deployment_id = get_deployment_id(deploy_parameters)
          @@logger.debug "Deployment id set to #{deployment_id}"
          coordinator_configuration = augment_with_coordination(deployment_id, deploy_parameters)
          # if a coordinator exists, coordinator_configuration will be defined
          # if it exists, we want to use the identifier assigned as part of establishing the coordination transaction
          # otherwise (no coordinator), create an id from scratch
          deployment_id = coordinator_configuration.nil? ? create_instance_id() : coordinator_configuration[:identifier]
          
          synchronous = properties[:synchronous].nil? ? false : properties[:synchronous]
          if synchronous and coordinator_configuration.nil?
            # synchronous deployment requires coordinator to check for vm completion
            raise WeaverError.new("Synchronous deployment requires coordinator")
          end
          
    
          transaction = coordinator_configuration.nil? ? nil : coordinator_configuration[:transaction]    

          @@logger.info "Deploying #{@root.id.inspect} with properties: '#{properties.inspect}' and deploy parameters: '#{deploy_parameters.inspect}'."
          apply_deployment_parameters(deploy_parameters)
          @root.topologies.values.each do |topology|
            provisioner = topology.provisioner || default_provisioner()
            deployer = provisioner.deployer(properties, &@reporter)
            deployer.deploy(topology, deploy_parameters)
          end
          
          top = @root.topologies.values.first
          top.amd[:deployment] = Weaver::Core::Attribute.new(:deployment, top)
          top.amd[:deployment].value = deployment_id
          top.amd[:deployment].tags << :novalidate

    
          if synchronous # (checked above) and !coordinator_configuration.nil?
            # record configuration of deployment; then external queries for status/list can proceed
            # will update after final status is known 
            record_configuration(transaction, deploy_parameters)
            # record updated model if we can; do it now so that can query for status
            write_model(deployed_model)
            # we ignore nodes with tags because a volume is a Node with tag :storage in this version
            @root.flatten_children.select { |c| c.kind_of?(Weaver::Core::Node) && c.realized_by.empty? && c.tags.empty? }.each do |node|
              node_status = transaction.read_data("#{node.fully_qualified_id}.config_status") { |v| v =~ /configured/ or v =~ /failed/ }
            end # @root.flatten_chilren.values.each
          end # if synchronous
        } # Timeout::timeout
      rescue Timeout::Error => te
        @@logger.warn "Fatal error #{te.inspect} deploying '#{@root.topologies.values.first.name}'"
        timed_out = te
      end
      
      # record configuration of deployment if known 
      record_configuration(transaction, deploy_parameters)
      # record updated model if we can
      write_model(deployed_model)
        
      raise timed_out unless timed_out.nil?
      
      return self      
    end
    
    # Updates a deployed instance (or creates if not deployed)
    #
    # * *Args*    :
    #   - +instance_id+ -> deployed instance id if exists
    #   - +properties+ -> a list of properties that are needed to configure the underlying deployer.
    #                     This is implementation specific and may contain username, password, key files, etc.
    #   - +deploy_parameters+ -> deployment time parameters for the environment topology (e.g., :deps_file_url)
    #
    # * *Returns* : the update environment instance (type Environment)
    #
    # * *Raises* :
    #   - +WeaverError+ if an error occurs while deploying the environment
    #
    def deploy_or_update(change_template, template_config, instance_id, properties = {}, deploy_parameters = {})
      if topology.nil?
        raise WeaverError.new("No topology to deploy")
      end
      
      # Retrieve a model of the existing environment
      # TODO: change this to use the service
      deployed_env = nil
      deployed_model = instance_id.nil? ? nil : "#{@instance_dir}/#{instance_id}/model.weaver"
      if !instance_id.nil?
        raise WeaverError.new("Deployed instance model cannot be found") unless File.exists?(deployed_model)
        deployed_env = Weaver::Dsl::WeaverParser.new.parse(deployed_model)
      end
      raise WeaverError.new("Deployed instance model invalid") if deployed_env.nil? && !instance_id.nil?
            
      if deploy_parameters[:mock] == "true"
        deploy_parameters[:mock_coordination] = "true" if deploy_parameters[:mock_coordination].nil?
      end
      
      @@logger.info "Creating #{@root.id.inspect} with properties: '#{properties.inspect}' and deploy parameters: '#{deploy_parameters.inspect}'."
      @@logger.info "Using change template '#{change_template}' configured with '#{template_config.inspect}'"

      coordinator_configuration = nil
      timed_out = nil
      transaction = nil 
      begin
        @@logger.debug "timeout set to #{properties[:timeout]}"
        Timeout::timeout(properties[:timeout]) {
          # apply deployment parameter to model
          apply_deployment_parameters(deploy_parameters)
          
          # Create action plan
          plan = Weaver::Execution::ChangePlan.new(change_template, template_config, @root, deployed_env)
          deployment_id = instance_id || create_instance_id()
          coordination_id = deployment_id # TODO: create_instance_id()
          deploy_parameters.merge!(:coordination_id => coordination_id)
          @@logger.debug "Executing plan:"
          @@logger.debug "#{plan}"
          
          # Execute plan creating an updated model
          @root = plan.execute(properties, deploy_parameters)
          @@logger.debug "Plan execution complete"
          
          transaction = plan.transaction
          
          # Update topology model with new deployment id
          top = @root.topologies.values.first
          @@logger.debug "Setting topology id to: #{deployment_id}" # #{coordinator_configuration[:identifier]}"
          top.amd[:deployment] = Weaver::Core::Attribute.new(:deployment, top)
          top.amd[:deployment].value = deployment_id # coordinator_configuration[:identifier]
          top.amd[:deployment].tags << :novalidate
    
          # If synchronous call, wait (via coordinator) for completion of deployment
          synchronous = properties[:synchronous].nil? ? false : properties[:synchronous]
          if synchronous and !coordinator_configuration.nil?
            # we ignore nodes with tags because a volume is a Node with tag :storage in this version
            @root.flatten_children.select { |c| c.kind_of?(Weaver::Core::Node) && c.realized_by.empty? && c.tags.empty? }.each do |node|
              node_status = transaction.read_data("#{node.fully_qualified_id}.config_status", :wait => false) { |v| v =~ /configured/ or v =~ /failed/ }
            end # @root.flatten_chilren.values.each
          end # if synchronous
        } # Timeout::timeout
      rescue Timeout::Error => te
        @@logger.warn "Fatal error #{te.inspect} deploying '#{@root.topologies.values.first.name}'"
        timed_out = te
      end

      # record configuration of deployment if known      
      record_configuration(transaction, deploy_parameters)

      # record updated model if we can
      write_model (deployed_model)
        
      raise timed_out unless timed_out.nil?
      
      return self      
    end

    def find_matching_node
      if Unroll.template?(request_node) then
        # the deployed instance has only 1 node deployed for this group
        # just map it to the first in the list
        request_node = Unroll.group(request_node).members.values.first
      end
    end
    
    def default_provisioner
      return Weaver::Providers::DelegatingProvider.new()
    end
    
    # Simulate the results of deployment by producing a model of the expected deployment.
    #
    # This method is mostly used for validation; for example it may delegate to a cloud-specific
    # deployer which will check static IP address requests and propose dynamic IP addresses that
    # are good enough that we can use them for validation.
    #
    # * *Args*    :
    #   - +properties+ -> a list of properties that are needed to configure the underlying deployer.
    #                     This is implementation specific and may contain username, password, key files, etc.
    #   - +deploy_parameters+ -> deployment time parameters for the environment topology (e.g., :deps_file_url)
    # * *Returns* :
    #   -+Environment+ The updated environment instance with runtime information such as hostname and ip.
    #
    # * *Returns* : a simulated post-deployment Environment
    #
    def simulate_deploy(properties = {}, deploy_parameters = {}, coordinator = nil)       
      @@logger.info "Simulating deployment of #{@root.id.inspect} with properties: '#{properties.inspect}' and deploy parameters: '#{deploy_parameters.inspect}'."
      copy_e = self.clone
      deployment_id = copy_e.get_deployment_id(deploy_parameters)
      @@logger.debug "Simulate identified deployment_id of #{deployment_id}"
      deploy_parameters.merge!({:coordinator => coordinator}) unless coordinator.nil?
      copy_e.augment_with_coordination(deployment_id, deploy_parameters)
      copy_e.apply_deployment_parameters(deploy_parameters)
      
      copy_e.root.topologies.values.each do |topology|
        provisioner = topology.provisioner || default_provisioner()
        deployer = provisioner.deployer({ :simulate=>true }.merge!(properties), &@reporter)
        deployer.simulate_deploy(topology, deploy_parameters)
      end
      
      top = copy_e.root.topologies.values.first
      top.amd[:deployment] = Weaver::Core::Attribute.new(:deployment, top)
      top.amd[:deployment].value = deployment_id
      top.amd[:deployment].tags << :novalidate

      return copy_e 
    end
    
    # Legacy support for CLI and Rakefiles that assume one topology
    # TODO deprecate
    def topology
      return @root.topologies.values.to_a[0]
    end

    # Serialize the current state of an environment topology and all its referenced topologies
    # to a Weaver file. The goal of the serialization is to redeploy that topology with the 
    # same configuration values as the original one (except for any IP addresses and hostnames).
    #
    # * *Args*    :
    #   - +filename+ -> the filename (type String) where to serialize it (e.g., "/tmp/xyz.weaver")
    #
    # * *Raises* :
    #   - +WeaverError+ if an error occurs while serializing the environment
    #
    def serialize(filename)
      # duplicate and call out to the deployer(s) to see whether they want to postprocess
      
      duplicate = Weaver::Transforms::Duplicate.duplicate(@root)
      
      # Remove imports, which have already been processed
      new_root = Root.new(duplicate.id)
      new_root.topologies.merge!(duplicate.topologies)
      duplicate = new_root
      
      duplicate.flatten_children.select { |child| child.kind_of?(Topology) }.each do |top|
        unless top.provisioner.nil?
          type = top.provisioner.provider_type  
          deployer = Weaver::Deployers::Deployer.find(type)
          if deployer.respond_to?(:serialize_topology_preprocessor)
            deployer.serialize_topology_preprocessor(top)
          end
        end
      end
      
      # serialize it to file  
      begin        
        File.open(filename, "w") do |f|
          f.write duplicate.to_weaver
        end
      rescue => e
        raise WeaverError, e.message, e
      end    
    end

    # Let developers known which Weaver things can be manipulated.
    def respond_to?(sym, include_private=false)
      # Allow the first inner Topology to be manipulated as the environment.      
      topology = @root.topologies.values.to_a[0]
      
      # NOTE: In Ruby 1.9.x it is better to define respond_to_missing?() than respond_to?(); we do this to be compatible.
      return super || (!topology.nil? && topology.respond_to?(sym, include_private))
    end

    def method_missing(sym, *arguments, &block)
      # Allow the first inner Topology to be manipulated as the environment.      
      topology = @root.nil? ? nil : @root.topologies.values.to_a[0]
      # this part is copied from Base.method_missing to have dynamic method on attributes
      if topology.nil?
        @@logger.warn "Weaver cannot evaluate #{sym} because there is no Weaver environment topology."
        return super
      end
      
      topology.send(sym, *arguments, &block)
    end

    protected
    
    # Retrieve deployment identifier; these methods could use some cleanup
    def get_deployment_id (deploy_parameters)
      deployment_id = deploy_parameters.has_key?(:coordination_id) ? "#{(root.id)[/[^\/]*$/]}\##{deploy_parameters[:coordination_id]}" : nil
      deployment_id = deploy_parameters.has_key?(:coordination_id) ? "#{deploy_parameters[:coordination_id]}" : nil
      # use a value we already have
      deployment_id = @root.topologies.values.first.deployment if !@root.topologies.empty? && @root.topologies.values.first.amd.include?(:deployment)
      # deployment_id = create_instance_id() if deployment_id.nil?
      deployment_id
    end
    
    def augment_with_coordination(instance_id, deploy_parameters) 
#      coordinator_hosts = Weaver::System.instance.get_property(:zookeeper_hosts)
#      return nil if coordinator_hosts.nil?

      Weaver::Coordination::System.instance.set_property(:log_level, Weaver::System.instance.get_property(:log_level))
#      Weaver::Coordination::System.instance.set_property(:zookeeper_hosts, coordinator_hosts)
#      Weaver::Coordination::System.instance.set_property(:zookeeper_port, (Weaver::System.instance.get_property(:zookeeper_port) || 2181))
#      Weaver::Coordination::System.instance.set_property(:transaction_prefix, (self.id)[/[^\/]*$/]) unless topology.nil?

#      instance_id = deploy_parameters.has_key?(:coordination_id) ? "#{(root.id)[/[^\/]*$/]}\##{deploy_parameters[:coordination_id]}" : nil
#      instance_id = deploy_parameters.has_key?(:coordination_id) ? "#{deploy_parameters[:coordination_id]}" : nil
#      # use a value we already have
#      instance_id = @root.topologies.values.first.deployment if !@root.topologies.empty? && @root.topologies.values.first.amd.include?(:deployment)

      transaction = nil;
      coordinator = deploy_parameters[:coordinator]
      coordinator = Weaver::Deployers::Execution::CoordinationUtil.get_coordinator if coordinator.nil?
      if instance_id.nil?
        @@logger.debug "Creating (new) coordination transaction"
        prefix = topology.nil? ? nil : (self.id)[/[^\/]*$/]
        @@logger.debug "prefix is '#{prefix}'"
        transaction = coordinator.create_transaction({:transaction_prefix => prefix})
      else
        @@logger.debug "Joining coordination transaction with instance_id = '#{instance_id}'"
        transaction = coordinator.join_transaction(instance_id)
      end
      @@logger.debug "Created configuration transaction #{transaction.inspect}"

      # quit now if the coordination is mocked
      # TODO fix this so that can complete the method
      # TODO be sure to use this method everywhere
      return unless Weaver::Coordination::CoordinatedTransactionManager.instance.can_coordinate?(transaction.coordinator)
      
      # compute dependencies and add extra parameters
      dependencies = Weaver::Deployers::Execution::CoordinationUtil.get_dependencies(@root)
      deploy_parameters.merge!(:weaver_coordination_topology_id => transaction.id)
      deploy_parameters.merge!(:weaver_coordination_dependencies => dependencies)
      
      # Record details from the model
      Weaver::Deployers::Execution::CoordinationUtil.write_defaults(transaction, root)

      # new logic for "all" in late bindings should make this unecessary      
#      # Some of the source values for dependencies are in the model. Write them to the coordinator now.
#      Weaver::Deployers::Execution::CoordinationUtil.write_constants(transaction, dependencies)
      
      return {
        :identifier => transaction.id,
        :transaction => transaction,
        :dependencies => dependencies
      }
    end
    
    # Fill in what configuration we can in the model and save the model 
    def record_configuration(transaction, deploy_parameters)
      # record configuration status as it is currently known
      
      mock_coordination = deploy_parameters[:mock_coordination] == "true"
        
      root_status = ConfigStatus::UNKNOWN

      # we ignore nodes with tags because a volume is a Node with tag :storage in this version
      nodes = @root.flatten_children.select { |c| c.kind_of?(Weaver::Core::Node) and c.realized_by.empty? and c.scope.kind_of?(Weaver::Core::Topology) and !Weaver::Transforms::Unroll.template?(c) and !c.tags.include?(:storage) }
      status_cache = {} 
      nodes.each do |node|
        # configuration status
        # default value is unknown
        node_status = Weaver::Coordination::CONFIG_STATUS::UNKNOWN
        # if coordination is enabled, update from coordinator
        unless transaction.nil? || mock_coordination
          node_status = transaction.read_data("#{node.fully_qualified_id}.config_status", :wait => false) ||  Weaver::Coordination::CONFIG_STATUS::UNKNOWN
        end
        @@logger.debug "status of #{node.fully_qualified_id} = #{node_status.inspect}"
        status_cache[node] = node_status
        node.amd[:config_status] = Weaver::Core::Attribute.new(:config_status, node)
        node.amd[:config_status].value = node_status

        # Use the node status to compute the resulting topology status             
        # However, ignore a node status of 'deleting' or 'deleted'; they are handled separately below 
        unless /^delet/ =~ node_status
          root_status = ConfigStatus.find_by_value(node_status) if root_status < ConfigStatus.find_by_value(node_status)
        end 
      end # @root.flatten_children.values.each
       
      # Now check if the topology is actually being deleted.
      # Pay attention only if _all_ nodes are deleting or deleted
      all_del = true
      del_status = ConfigStatus::UNKNOWN
      status_cache.each do |node, node_status|
        if /^delet/ =~ node_status
          del_status = ConfigStatus.find_by_value(node_status) if del_status < ConfigStatus.find_by_value(node_status)
          @@logger.debug "set del status to #{del_status.inspect}"
        else
          all_del = false
          break
        end
      end
      @@logger.debug "all_del = #{all_del}; del_status = #{del_status.inspect}"
      if all_del
        root_status = del_status 
      end
         
       # Record a status for the deployment as a whole
       top = @root.topologies.values.first
       top.amd[:config_status] = Weaver::Core::Attribute.new(:config_status, top) if top.amd[:config_status].nil?
       top.amd[:config_status].value = root_status.label
         
     end # def record_configuration

    # record updated model if we can
    # TODO: replace with call to configuration database
    def write_model (location)
      location = "#{@instance_dir}/#{@root.topologies.values.first.deployment}/model.weaver" if location.nil? and @root.topologies.values.first.amd.include?(:deployment)
      unless location.nil? then  
        FileUtils.mkdir_p(File.dirname(location)) unless File.directory?(File.dirname(location))
        file = File.open(location, 'w')
        File.open(file.path, 'w') do |f|
          f.write(@root.to_weaver)
        end # 
      end 
    end # def write_model
    
    # Retrieve the most recent coordination id for a model; nil if none
    def last_coordination_id
      begin
        File.new("#{@instance_dir}/coord", 'r'); f.inject([]) {|r,op| r << op.gsub(/\n$/,''); r}.last
      rescue Exception => e
        nil
      end
    end # def last_coordination_id
    
    # Append a new coordination_id to the record of coordination ids for a model
    # A coordination id will be written only if there really is a coordinator available
    def write_coordination_id (coordination_id)
      # if there is no coordinator, then the coordination_id is for a simulated envrionment
      if Weaver::Coordination::CoordinatedTransactionManager.instance.coordinator?
        return unless simluated
      end

      File.open("#{@instance_dir}/coord", 'a') do |f|
        f << coordination_id
      end
    end # def write_coordination_id
     
    def create_instance_id ()
      # Generate a coordination transaction identifier; currently used as the deployment identifier
      instance_id = (self.id)[/[^\/]*$/]
      instance_id << "#" unless instance_id.empty?
      instance_id << SecureRandom.uuid
      
      return instance_id
    end # def create_instance_id

    # Inject deployment parameters into all Core::Attribute s with matching IDs
    # +deploy_parameters+ Hash of ID -> value, typically <code>{ :deps_file_url => 'http://example.ibm.com/...' }</code>
    def apply_deployment_parameters(deploy_parameters)
      @@logger.debug "Parsing deploy parameters in apply_deployment_parameters: #{deploy_parameters.inspect}"
      
      # stick the :deps_file_url into the model
      if deploy_parameters.include?(:deps_file_url) && deploy_parameters[:deps_file_url] != nil
        Weaver::Chef::ChefAutomationHandler.new().inject_deps_file_url(self, deploy_parameters[:deps_file_url])
        @@logger.debug "Injecting :deps_file_url => #{deploy_parameters[:deps_file_url]} into chef_automation statements"
      else 
        @@logger.debug "Chose to ignore :deps_file_url because: deploy_parameters.include?(:deps_file_url): #{deploy_parameters.include?(:deps_file_url)} or deploy_parameters[:deps_file_url] != nil: #{deploy_parameters[:deps_file_url] != nil}"
      end
      
      deploy_parameters.each do |k, v|
        if(k != :deps_file_url) 
          @root.flatten_attributes.select { |a| a.id == k }.each do |attr|
            @@logger.debug "Setting deployment attribute: #{k} = #{v}"
            attr.value = v
          end
        end
      end
      
    end

  end
end