require 'timeout'
require 'weaver/core/logging'
require 'weaver/core/weaver_error'
require 'weaver/coordination/coordinated_transaction_manager'
require 'weaver/execution/actions/action'

module Weaver
  module Execution
    module Actions
    
      # Waits for a set of actions to complete on a set of nodes.
      # An input implies no wait
      
      # This code can be called either by the Weaver service or by an automation executing on a VM
      # When called by the Weaver service, the option :coordinators need not be set
      # When called by an automation, the :coodinators option should be explicitly set
      
      # TODO: how to deal with logging class?
      
      class WaitAction < Weaver::Execution::Actions::Action
        
        DEFAULT_DELAY = 0
        DEFAULT_TOTAL_TIME = 180 # seconds
        
        # valid options include :delay
        def initialize (wait_map = {}, options = {})
          @wait_map = wait_map
          @delay = options[:delay] || DEFAULT_DELAY
          @timeout = options[:timeout] || DEFAULT_TOTAL_TIME
          @delete_wait = options[:delete] || false
        end # end def initialize
                      
        # TODO figure out how this works for WCC
        def execute  (properties, execute_parameters = {})
          raise Weaver::Core::WeaverError.new("Coordination transaction identifier required.") if execute_parameters[:coordination_id].nil?

          # wait for specified actions to complete
          wait(execute_parameters)
          
          # if :delay has been set, wait specified time
          @@logger.debug "WaitAction sleeping for #{@delay} seconds"
          sleep (@delay)
        end # end def execute
        
        # returns true if identifies a successful completion for all actions
        # returns failed if any action is "failed" _or_ the overall time is execeeded
        def wait (execute_parameters = {})
          hosts = execute_parameters[:coordinators]
          if hosts.nil?
            # should only happen on the weaver service
            require 'weaver/system'
            hosts = Weaver::System.instance.get_property(:zookeeper_hosts)
          end

          begin
            # short circuit if nothing to wait for
            return if @wait_map.empty?
            
            Timeout::timeout(@timeout) {
              tx = Weaver::Coordination::CoordinatedTransactionManager.instance.join_transaction(execute_parameters[:coordination_id])

              @wait_map.each do |node, ops|
                if ops.empty? or @delete_wait then
                  # wait on node status
                  @@logger.debug "Waiting for completion of node configuration for: '#{node}'"
                  nd_status = tx.read_data("#{node.fully_qualified_id}.config_status") { |status| ['configured', 'failed', 'deleted'].include?(status) }
                  return false if nd_status == "failed"
                  return false if nd_status == 'deleted' and !@delete_wait
                else
                  # wait on specific operation status
                  ops.each do |operation|
                    @@logger.debug "Waiting for completion of operation configuration for: '#{node}.#{operation}'"
                    op_status = tx.read_data("#{operation.fully_qualified_id}.config_status") { |status| ['configured', 'failed', 'deleted'].include?(status) }
                    return false if op_status == "failed"
                  end # ops.each do
                end # if ops.empty?
              end # @node_map.each do
              
              return true
              
            } # Timeout::timeout
          rescue Timeout::Error => te
            @@logger.warn "Fatal error #{te.inspect} waiting for actions to complete '#{@root.topologies.values.first.name}'"
            return false
          end # begin.. rescue.. end
        end # def wait
        
        def to_s
          retval = "WaitAction["
          retval << "(:delay => #{@delay}, :delete => #{@delete_wait})"
          retval << @wait_map.map { |node, automations| "   #{node.fully_qualified_id}[#{automations.map{|a| a.id.to_s}.join(',')}]" }.join(';')
          retval << "]"
          
          return retval
        end # def to_s
        
      end # class DeleteNodesAction
    
    end # module Action
  end # end module Execution
end # mmodule Weaver