require 'weaver/dsl/weaver_parser'

module Weaver
  module Generators
    
    # The EnvironmentGenerator generates skeleton environment file to act as a starting
    # point for the DevOps folks. It will provide realization and contain all required attributes 
    # that need to be set before the deployment of an environment.
    #
    # The Weaver CLI uses this class to perform the environment generation.
    #
    class EnvironmentGenerator
      include Weaver::Core
      
      # attributes that we do not augment even if invalid
      ACCEPTED_ATTRIBUTES = Set.new [ :chef_node, :chef_node_attr, :deps_file_url ]

      # Creates a new EnvironmentGenerator
      #   - +application_def_file+ -> optional filename of the application definition
      #   - +infrastructure_def_file+ -> the filename of the infrastructure definition
      # * *Returns* :
      #   - An EnvironmentGenerator for those two files.
      def initialize(application_def_file, infrastructure_def_file)
        @parser = Weaver::Dsl::WeaverParser.new()
        @application_def_file = application_def_file
        @infrastructure_def_file = infrastructure_def_file
        @infrastructure = @parser.parse(infrastructure_def_file).infrastructure
        if application_def_file != nil
          @application = @parser.parse(application_def_file).app       
        end
      end

      # Generates the environment skeleton and stores it in the given file.
      # * *Args*    :
      #   - +filename+ -> the filename for the generated environment definition (String)
      #   - +env_id+ -> the environment id to use (Symbol)
      #   - +env_name+ -> the environment name (Sting; optional)
      #   - +env_description+ -> the environment name  (String; optional)
      #   - +realizations+ -> a Hash defining the realization from logical app nodes to infrastructure nodes (Hash; optional)
      #   - +generate_defaults+ -> a boolean indicated whether defaults attribute values should be copied from application and infrastructure definitions
      # * *Returns* :
      #   - The content of the file
      def generate_environment_skeleton(filename, env_id, env_name = 'TODO', env_description = 'TODO', realizations = nil, generate_defaults = false)
        # container to hold imports
        root = @parser.parse(nil)
        
        unless @application.nil?
          import_app = Import.new(nil, root)
          @application.scope = import_app
          import_app.topologies[@application.id] = @application
          root.import(nil, import_app, @application_def_file)
        end
        import_inf = Import.new(nil, root)
        @infrastructure.scope = import_inf
        import_inf.topologies[@infrastructure.id] = @infrastructure
        root.import(nil, import_inf, @infrastructure_def_file)
        
        if env_id.kind_of?(String)
        	env_id =	env_id.symbolize
       	end
        env = Topology.new(env_id, root)
        root.topologies[env_id] = env
        root.prefix_comments.push ""
        root.prefix_comments.push "Generated automatically by weaver on #{DateTime.now.strftime('%FT%R')}"
        root.prefix_comments.push ""
        if @application.nil?
          env.prefix_comments.push ":#{env_id} represents a configuration of virtual infrastructure :#{@infrastructure.id}"
        else
          env.prefix_comments.push ":#{env_id} represents a configuration of :#{@application.id}" 
          env.prefix_comments.push "  realized upon virtual infrastructure :#{@infrastructure.id}"
        end
        env.name = env_name || 'TODO'
        env.description = env_description || 'TODO'
        
        build_tag = Attribute.new(:build_tag, env, true)
        build_tag.description = "Build tag from RTC"
        build_tag.value = "<to_be_defined>"
        build_tag.prefix_comments.push "define environment specific properties (values will be \"set/overridden\" from outside -> Rakefile)" 
        env.amd[build_tag.id] = build_tag
        
        #chef_cookbook_dep_url = Attribute.new(:chef_cookbook_dep_url, env, true) 
        #chef_cookbook_dep_url.description = "Dependency file URL for Chef cookbooks"
        #env.amd[chef_cookbook_dep_url.id] = chef_cookbook_dep_url
        
        env.unresolved_refinements[@application.id] = import_app.topologies[@application.id] unless @application.nil?
        env.unresolved_refinements[@infrastructure.id] = import_inf.topologies[@infrastructure.id]
        
#        env.topologies[@application.id] = @application if @application
#        env.topologies[@infrastructure.id] = @infrastructure

        if realizations.nil?
          unless @application.nil?
            @application.nodes.each do |id, node|
              env.unresolved_realizations[node.fully_qualified_id] = 'todo'
            end
          end
        else
          realizations.each do |k, v|
            raise WeaverError.new("Provided realizations not valid. Cannot find application node '#{k.to_sym}'.") unless @application.nodes.key?(k)
            raise WeaverError.new("Provided realizations not valid. Cannot find infrastructure node '#{v.to_sym}'.") unless @infrastructure.nodes.key?(v)
            env.unresolved_realizations[@application.nodes[k].fully_qualified_id] = @infrastructure.nodes[v].fully_qualified_id
          end
        end
        
        # Generate a comment for all application attributes
        unless @application.nil?
          # Suggest to user that that should import some Chef roles
          @application.nodes.each { |id, node| env.unresolved_additional_declarations.push "#{node.fully_qualified_id}.include 'automation/todo_role.weaver'" }
          
          # Fake IP addresses so that validation doesn't complain about attributes that propagate IP addresses from Nodes
          @application.nodes.each { |id, node| node.ip_address = '127.0.0.1' }
            
          # Determine which attributes need values
          invalid_app_attrs = Set.new
          
          @application.flatten_attributes.each do |attr|
            if attr.mutable && !ACCEPTED_ATTRIBUTES.member?(attr.id)
              msgs = []
              attr.validate_collect(msgs, [:predeployment])
              unless msgs.empty?
                if invalid_app_attrs.empty?
                  env.unresolved_additional_declarations.push ''
                  env.unresolved_additional_declarations.push '# The following application properties require values'
                end
                
                env.unresolved_additional_declarations.push("#{attr.fully_qualified_id} 'TODO'")
                invalid_app_attrs.add(attr)
              end
            end
          end
          
#          env.unresolved_additional_declarations.push ''
#          env.unresolved_additional_declarations.push 'Uncomment application properties for customization'
#          @application.flatten_attributes.each do |attr|
#            if attr.mutable and !attr.implicit? && !invalid_app_attrs.member?(attr)
#              env.unresolved_additional_declarations.push("#{attr.fully_qualified_id} = 'TODO'")
#            end
#          end
        end
        
        # Generate a comment for all uncovered invalid infrastructure attributes
        reverse_realizations = realizations.nil? ? {} : realizations.invert
        invalid_inf_attrs = Set.new
        valid_inf_attrs = Set.new
        @infrastructure.flatten_attributes.each do |attr|
          if attr.mutable && !ACCEPTED_ATTRIBUTES.member?(attr.id)
            # We have a mutable attribute, but we don't want to create a TODO if it will be replaced by an App Node attribute
            app_node = @application.nil? ? nil : @application.nodes[reverse_realizations[attr.scope.id]]
            if app_node.nil? || app_node.amd[attr.id].nil?
              # It will not be replaced, is it invalid as it stands?
              msgs = []
              attr.validate_collect(msgs, [:predeployment])
              if msgs.empty?
                valid_inf_attrs.add(attr)
              else
                if invalid_inf_attrs.empty?
                  env.unresolved_additional_declarations.push ''
                  env.unresolved_additional_declarations.push '# The following infrastructure properties require values'
                end
                
                env.unresolved_additional_declarations.push("#{attr.fully_qualified_id} = 'TODO'")
                invalid_inf_attrs.add(attr)
              end
            end
          end
        end
        
#        env.unresolved_additional_declarations.push ''
#        env.unresolved_additional_declarations.push 'Uncomment infrastructure properties for customization'
#        valid_inf_attrs.each do |attr| 
#          unless attr.implicit?
#            env.unresolved_additional_declarations.push("#{attr.fully_qualified_id} = 'TODO'")
#          end
#        end
        
        content = root.to_weaver
        File.open(filename, 'w') do |f|
          f.puts content
        end
        return content
      end

      private

    end
  end
end
