module Weaver
  # \Weaver::Generators provides tools to generate \Weaver files for various roles
  #
  # @see EclipseAppGenerator
  # @see EnvironmentGenerator
  # @see InfFacadeGenerator
  module Generators
    # The InfFacadeGenerator generates a Weaver facade from an infrastructure.
    class InfFacadeGenerator
      include Weaver::Core
      
      # * *Args*    :
      #   - +inf+ -> the id of the facade Infrastructure
      #   - +inf_filename+ -> the filename of the infrastructure file (the generated facade will import it by name)
      def initialize(inf, inf_filename)
        @inf = inf.clone
        @inf_filename = inf_filename
      end

      # Generates a facade
      # * *Args*    :
      #   - +facade_inf_id+ -> the id of the facade Infrastructure
      #   - +filename+ -> the filename to store the file (String; optional).
      #                   If not given the application definition is not saved
      #   - +mapping+ -> defines a node to component mapping (optional, Hash with id -> id)
      #
      # * *Returns* :
      #   - the generated Infrastructure model
      #
      def generate_inf_facade(facade_inf_id, filename, mapping = {})
        facade = Root.new(filename, nil)
        facade.prefix_comments.push ""
        facade.prefix_comments.push "Generated automatically by weaver on #{DateTime.now.strftime('%FT%R')}"
        facade.prefix_comments.push ""
        
        import = Import.new(:original, facade)
        @inf.scope = import
#        import.filespec = relativize(@inf_filename, File.dirname(filename)) 
        import.topologies[@inf.id] = @inf
        
        facade.import(:original, import, relativize(@inf_filename, File.dirname(filename)))
        
        facade.unresolved_refinements[facade_inf_id] = import.topologies[@inf.id]
 
        # We can't put the real infrastructure into the augmentation as it is fully specified,
        # just sketch out its type and inject some refinements
        sketch = Topology.new(facade_inf_id, facade)
        sketch.type = @inf.type
        sketch.scope = facade
        sketch.provenance = @inf.provenance
        facade.topologies[facade_inf_id] = sketch
        push_mappings(sketch, nil, mapping)
        
        @@logger.debug "infrastructure renamings: #{sketch.unresolved_renamings.inspect}"

        unless filename.nil?
          # write the facade model to a file
          file = File.new(filename, 'w')
          file.write(facade.to_weaver)
          file.close
        end

        return facade
      end

      # Given an infrastructure, add renames not supplied
      # * *Args*    :
      #   - +inf+ -> any NodeContainerMixin
      #   - +raw_renamings+ -> Hash of Node to Symbol which will receive new entries
      def ensure_all_nodes_renamed(inf, raw_renamings)
        n = 0
        inf.nodes.each do |id, node| 
          unless raw_renamings.member?(id)
            raw_renamings[id] = ("nodename_TODO_" + n.to_s).to_sym
            n = n+1
          end
        end
      end

      # Given a Hash +raw_renamings+, build up a tree form matching the model        
      # * *Args*    :
      #   - +entity+ -> any Base
      #   - +raw_renamings+ -> Hash of Base to Symbol
      # RECURSIVE
      def build_tree_form(entity, raw_renamings)
        tree_form = {}
        entity.child_entities.each do |id, child|
          # RECURSIVE
          child_renamings = build_tree_form(child, raw_renamings)
          if child_renamings.empty?
            tree_form[id] = raw_renamings[id] if raw_renamings.has_key?(id) 
          else
            tree_form[id] = Array.[](raw_renamings[id], child_renamings) 
          end
        end
        return tree_form
      end
      
      # Return the filename relative to 'dir'
      def relativize(filename, dir)
        dirname = File.expand_path(dir) + '/'
        prefix = ''
        while dirname
          dirname = '/' if dirname == '//'
          startswith = Regexp.new("^#{dirname}(.*)").match(filename)
          #puts "dirname=#{dirname}, prefix=#{prefix}, startswith=#{startswith}"
          return "#{prefix}#{startswith[1]}" if startswith
          dirname = dirname == '/' ? nil : (File.dirname(dirname) + '/')
          prefix = "../#{prefix}"
        end
        raise WeaverError.new("Couldn't handle '#{dir}' (which expands to '#{File.expand_path(dir)}') for '#{filename}'")
      end
            
      protected

      # Recursively creates renamings from 'mapping' and pushes them into 'entity's @unresolved_renamings
      # * *Args*    :
      #   - +entity+ -> the entity to hold the renamings
      #   - +prefix+ -> prefix to be applied to all maps
      #   - +mapping+ -> hash of Symbol id to either Symbol id or [Symbol id, hash-of-child mappings]
      #
      # * *Returns* :
      #   - the generated Infrastructure model
      #
      def push_mappings(entity, prefix, mapping)
        mapping.each do |source_id, id_or_array|
          if id_or_array.kind_of?(Array)
            raise WeaverError.new("Bad mapping; invalid array") unless id_or_array.length == 2
            target_id = id_or_array[0]
            target_child_mappings = id_or_array[1]
            raise WeaverError.new("Bad mapping; invalid child mappings #{target_child_mappings.class}") unless target_child_mappings.kind_of?(Hash)
          else
            target_id = id_or_array
            target_child_mappings = nil
          end
          raise WeaverError.new("Bad mapping; invalid source_id #{source_id.class}") unless source_id.kind_of?(Symbol)
          raise WeaverError.new("Bad mapping; invalid target_id #{target_id.class}") unless target_id.kind_of?(Symbol) || target_id.nil?
          
          entity.unresolved_renamings.push({ "#{prefix}#{'.' unless prefix.nil?}#{source_id}" => target_id }
            ) unless target_id.nil? || target_id == source_id 
          
          unless target_child_mappings.nil?
            # RECURSIVE
            push_mappings(entity, "#{prefix}#{'.' unless prefix.nil?}#{target_id || source_id}", target_child_mappings)
          end
        end
      end

    end

  end

end
