require 'weaver/system'
require 'weaver/core/status_constants'
require 'yaml'

module Weaver
  # Contains all IWD specific code
  module Iwd
    # Util methods for IWD weaver patterns/infrastructures
    #
    #
    class PatternUtils
      include Weaver::Core
      
      # Return the first iwd_node (Node) that has the given part id, or nil if it's not there
      #
      # * *Args*    :
      #   - +iwd_pattern+ -> the IWD pattern (Infrastructure)
      #   - +part_id+ -> the part id that is being requested (Fixnum)
      # * *Returns* :
      #   - An iwd_node (Node) or nil
      #
      def self.get_part_from_pattern(iwd_pattern, part_id)
        return get_part(iwd_pattern.nodes, part_id)
      end
      
      # Return a Hash of the original list of nodes, without aliased nodes
      #
      # * *Args*    :
      #   - +iwd_pattern+ -> the IWD pattern (Infrastructure)
      # * *Returns* :
      #   - A  hash of original Nodes
      #
      def self.get_unrealized_nodes(iwd_pattern)
        # The expression asks @current to flatten it its unique direct-and-indirect children (so if a Topology contains a Topology
        # those nodes are included).  We select only for unrealized nodes (this shouldn't apply to our current generated IWD so
        # the 'realized_by.nil?' doesn't really do anything.)  We take the nodes and build [id, node] pairs and construct a Hash from them.
        return Hash[iwd_pattern.flatten_children.select { |child| child.kind_of?(Node) && child.realized_by.empty? }.map {|child| [child.id, child] }]
      end
      
      # Return a list of the original list of nodes, without aliased nodes
      #
      # * *Args*    :
      #   - +iwd_pattern+ -> the IWD pattern (Infrastructure)
      # * *Returns* :
      #   - A  list of original Nodes
      #
      def self.get_unrealized_nodes_list(iwd_pattern)
        # The expression asks @current to flatten it its unique direct-and-indirect children (so if a Topology contains a Topology
        # those nodes are included).  We select only for unrealized nodes (this shouldn't apply to our current generated IWD so
        # the 'realized_by.nil?' doesn't really do anything.) 
        return iwd_pattern.flatten_children.select { |child| child.kind_of?(Node) && child.realized_by.empty? }
      end

      # Return the first iwd_node (Node) that has the given part id, or nil if it's not there
      #
      # * *Args*    :
      #   - +iwd_nodes+ -> a Hash of iwd_nodes(Nodes)
      #   - +part_id+ -> the part id that is being requested (Fixnum)
      # * *Returns* :
      #   - An iwd_node (Node) or nil
      #
      def self.get_part(iwd_nodes, part_id)
        unless (iwd_nodes.nil?)
          iwd_nodes.each do | k, v |
            if (v.part_id == part_id)
              return v
            end
          end
        end
        return nil
      end

      # Return the key to the first iwd_node that has the given part id, or nil if it's not there
      #
      # * *Args*    :
      #   - +iwd_nodes+ -> a Hash of iwd_nodes(Nodes)
      #   - +part_id+ -> the part id that is being requested (Fixnum)
      # * *Returns* :
      #   - An iwd_node (Node) or nil
      #
      def self.get_part_key(iwd_nodes, part_id)
        unless (iwd_nodes.nil?)
          iwd_nodes.each do | k, v |
            if (v.part_id == part_id)
              return k
            end
          end
        end
        return nil
      end
      
      # Given a hash of Nodes, return the key for the node that has the given image id, or nil if it's not there
      #
      # * *Args*    :
      #   - +nodes+ -> a Hash of nodes (Node)
      #   - +image_id+ -> the image id that is being requested (String)
      # * *Returns* :
      #   - A node (Node) or nil
      #
      def self.get_image_id_key(nodes, image_id)
        unless (nodes.nil?)
          nodes.each do | k, v |
            if (v.image_id == image_id)
              return k
            end
          end
        end
        return nil
      end

      # Return true if the Hash of iwd_nodes (Node)  has a part with the given part id
      #
      # * *Args*    :
      #   - +iwd_nodes+ -> a Hash of iwd_nodes(Nodes)
      #   - +part_id+ -> the part id that is being requested (Fixnum)
      # * *Returns* :
      #   - true if the list includes the part
      #
      def self.has_part_id(iwd_nodes, part_id)
        unless (iwd_nodes.nil?)
          iwd_nodes.each do | k, v |
            if (v.part_id == part_id)
              return true
            end
          end
        end
        return false
      end

      # Given a hash of Automations, return the script package (Automation) that has the given script id, or nil if it's not there
      #
      # * *Args*    :
      #   - +automation+ -> a Hash of automations (Automation)
      #   - +script_id+ -> the script id that is being requested (Fixnum)
      # * *Returns* :
      #   - A script package (Automation) or nil
      #
      def self.get_script_package(automation, script_id)
        unless (automation.nil?)
          automation.each do | k, v |
            if v.type == :script_package
              if (v.script_id == script_id)
                return v
              end
            end
          end
        end
        return nil
      end

      # Given a hash of Automations, return the key for the script package that has the given script id, or nil if it's not there
      #
      # * *Args*    :
      #   - +automation+ -> a Hash of automations (Automation)
      #   - +script_id+ -> the script id that is being requested (Fixnum)
      # * *Returns* :
      #   - A script package (Automation) or nil
      #
      def self.get_script_package_key(automation, script_id)
        unless (automation.nil?)
          automation.each do | k, v |
            if v.type == :script_package
              if (v.script_id == script_id)
                return k
              end
            end
          end
        end
        return nil
      end
      
      # Given a hash of Attributes, return the key for the attribute that has the given name and group, or nil if it's not there
      #
      # * *Args*    :
      #   - +attributes+ -> a Hash of Attributes
      #   - +name+ -> the name to match
      #   - +group+ -> the group to match
      # * *Returns* :
      #   - An Attribute or nil
      #
      def self.get_property_key(attributes, name, group)
        unless (attributes.nil?)
          attributes.each do | k, v |
              if (v.name == name && v.group == group)
                return k
            end
          end
        end
        return nil
      end
      
      # Return the IWD connection properties
      # Looks for IWD properties in Weaver System variables first, iwd.hostname, iwd.user, iwd.passwod, iwd.use_ssl.
      # If any of the System properties are undefined, looks for them in a yml file
      # pointed to by ENV["PROPS_FILE"]
      #
      # * *Args*    :
      #   - +pattern+ -> the pattern on which to report errors
      #   - +reporter+ -> the reporter for errors, or nil
      # * *Returns* :
      #   - A Hash of IWD properties (hostname, user, password, ssl_verify_cert)
      #
      def self.get_iwd_properties(pattern, reporter = nil, is_predeployment = false)
        
        iwd_properties = nil
        hostname = username = password = ssl_verify_cert = nil
        
        # First look for values in the config parameters that are given with in config.yml of a Weaver project
        config = pattern.config
        if (config && config[:iwd_properties])
          hostname = config[:iwd_properties][:hostname] 
          username = config[:iwd_properties][:user]
          password = Weaver::Utils::ValueDecryptor.decrypt_if_needed("iwd_properties.password", config[:iwd_properties][:password])
          ssl_verify_cert = config[:iwd_properties][:ssl_verify_cert]
         end
        
        # Next look for values in Weaver System properties
        if (hostname.nil? || hostname.empty? ||
        username.nil? || username.empty? ||
        password.nil? || password.empty? ||
        ssl_verify_cert.nil? || (ssl_verify_cert.kind_of?(String) && ssl_verify_cert.empty?))
          hostname = Weaver::System.instance.get_property("iwd.hostname") if (hostname.nil? || hostname.empty?)
          username =  Weaver::System.instance.get_property("iwd.user") if (username.nil? || username.empty?)
          password = Weaver::Utils::ValueDecryptor.decrypt_if_needed("System Properties iwd.password", Weaver::System.instance.get_property("iwd.password")) if (password.nil? || password.empty?)
          ssl_verify_cert = Weaver::System.instance.get_property("iwd.use_ssl") if (ssl_verify_cert.nil? || ssl_verify_cert.empty?)
        end
        
        # Next look for values from a properties file
        if (hostname.nil? || hostname.empty? || 
          username.nil? || username.empty? || 
            password.nil? || password.empty? || 
              ssl_verify_cert.nil? || (ssl_verify_cert.kind_of?(String) && ssl_verify_cert.empty?))
          filename = ENV['PROPS_FILE']
          if (filename != nil && filename != "")
            if File.exists?(filename)
              begin
                inf_properties =  YAML::load_file(filename)
                if inf_properties && !inf_properties['iwd_properties'].nil?
                  hostname = inf_properties['iwd_properties']['hostname'] unless (!hostname.nil? && !hostname.empty?)
                  username = inf_properties['iwd_properties']['user'] unless (!username.nil? && !username.empty?)
                  password = Weaver::Utils::ValueDecryptor.decrypt_if_needed("PROPS_FILE iwd_properties.password", inf_properties['iwd_properties']['password']) unless (!password.nil? && !password.empty?)
                  if (ssl_verify_cert.nil? || (ssl_verify_cert.kind_of?(String) && ssl_verify_cert.empty?)) 
                    ssl_verify_cert = inf_properties['iwd_properties']['ssl_verify_cert']
                  end
                end
              rescue Exception => e
                msg = Weaver::IWDUtils::MESSAGES[:error_loading_yml_file][:label] % ["#{filename}", "#{e.class}"]
                sev_type = (is_predeployment ? :predeployment_severity : :severity)
                severity = Weaver::IWDUtils::MESSAGES[:error_loading_yml_file][sev_type]
                if !reporter.nil?
                  problem_status = reporter.report(severity, msg, pattern)
                  problem_status.exception = e
                  problem_status.add_resolution(Weaver::IWDUtils::RESOLUTION_LABELS[:set_props_file_var], Weaver::IWDUtils::MESSAGES[:set_props_file_var][:label] % ["PROPS_FILE"], Weaver::Core::ResolutionTypes::SET_CREDENTIALS)
                end
              end
            # if one of the credentials is nil, and PROPS_FILE points to a file that does not exist, print a message 
            # (don't print if ssl_verify_cert is nil, since that's a valid default)
            elsif (hostname.nil? || hostname.empty? || 
              username.nil? || username.empty? || 
                password.nil? || password.empty?)
              msg = Weaver::IWDUtils::MESSAGES[:props_file_does_not_exist][:label] % ["PROPS_FILE", "#{filename}"]
              sev_type = (is_predeployment ? :predeployment_severity : :severity)
              severity = Weaver::IWDUtils::MESSAGES[:props_file_does_not_exist][sev_type]
              if !reporter.nil?
                problem_status = reporter.report(severity, msg, pattern)
                problem_status.add_resolution(Weaver::IWDUtils::RESOLUTION_LABELS[:set_props_file_var], Weaver::IWDUtils::MESSAGES[:set_props_file_var][:label] % ["PROPS_FILE"], Weaver::Core::ResolutionTypes::SET_CREDENTIALS)
              end
            end
          end
        end

        unless ssl_verify_cert.nil? || !ssl_verify_cert.kind_of?(String)
          ssl_verify_cert = ssl_verify_cert.match(/(true|t|yes|y|1)$/i) != nil 
        end

        if (hostname.nil? || hostname.empty? || username.nil? || username.empty? || password.nil? || password.empty?)
          hidden_password = nil
          if !password.nil?
            hidden_password = password.gsub(/./,"*")
          end
          msg = Weaver::IWDUtils::MESSAGES[:invalid_credentials][:label] % ["#{hostname}","#{username}", "#{hidden_password}", "#{ssl_verify_cert}", "#{ENV['PROPS_FILE']}"]
          sev_type = (is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:invalid_credentials][sev_type]
          if !reporter.nil?
            problem_status = reporter.report(severity, msg, pattern)            
            problem_status.add_resolution(Weaver::IWDUtils::RESOLUTION_LABELS[:set_weaver_system_properties], Weaver::IWDUtils::MESSAGES[:set_weaver_system_properties][:label], Weaver::Core::ResolutionTypes::SET_CREDENTIALS)
            problem_status.add_resolution(Weaver::IWDUtils::RESOLUTION_LABELS[:set_env_var_props_file], Weaver::IWDUtils::MESSAGES[:set_env_var_props_file][:label], Weaver::Core::ResolutionTypes::SET_CREDENTIALS)
          end
        else
          iwd_properties = {}
          iwd_properties[:hostname] = hostname
          iwd_properties[:username] = username
          iwd_properties[:password] = password
          iwd_properties[:ssl_verify_cert] = ssl_verify_cert || false
        end
        return iwd_properties

      end
      
      # Adds the hostname and ipaddr macros to a node for enabling correct
      # IWD parameter propagation
      # * *Args*    :
      #   - +iwd_version+ ->  the version of the iwd server
      #   - +node+ -> the node to add the properties
      # * *Returns* :
      #   - The updated node (type Node)
      def self.add_iwd_macros_to_node(iwd_version, node)
        if node.amd[:part_name].nil?
          @@logger.warn("#{node.fully_qualified_id} has no part_name")
          return
        end
        
        part_id = node.amd[:part_name].value
        if (iwd_version == "3.1.0.1")
          node.hostname = "${#{part_id}_.hostname}.${#{part_id}_.domain}"
          node.ip_address = "${#{part_id}_.ipaddr}"
        else
          node.hostname = "${#{part_id}.hostname}.${#{part_id}.domain}"
          node.ip_address = "${#{part_id}.ipaddr}"
        end
        @@logger.debug("IWD macros are set to #{node.hostname} and #{node.ip_address}")
        return node
      end
      
      # Return the delta between two Time objects, a start Time and an end Time
      # Here is a sample of a call to this method:
      #     start_time = Time.now
      #     function
      #     end_time = Time.now
      #     puts ("elapsed time: #{Weaver::Iwd::PatternUtils.time_delta_in_ms(start_time, end_time)} milliseconds")
      #
      # * *Args*    :
      #   - +start_time+ -> a Time representing the start time
      #   - +end_time -> a Time representing the end time
      # * *Returns* :
      #   - A Float with the delta of between the start and end time in milliseconds
      #
      def self.time_delta_in_ms(start_time, end_time)
        start = start_time.to_f
        endd = end_time.to_f
        (endd - start) * 1000.0
      end
      
      def self.report_pattern_exception(e, pattern, iwd_properties, reporter = nil, is_predeployment = false, other_severity=Logger::INFO, other_message=nil)
        pattern_id = pattern.amd[:pattern_id].nil? ? "<none>" : pattern.amd[:pattern_id].value
        if !e.message.nil? && e.message.index("401")
          msg = Weaver::IWDUtils::MESSAGES[:unauthorized_access][:label] % ["#{iwd_properties[:hostname]}", "#{iwd_properties[:username]}"]
          sev_type = (is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:unauthorized_access][sev_type]
#          msg = msg + ", " + e.message
          res_description = Weaver::IWDUtils::MESSAGES[:change_credentials][:label] % ["#{iwd_properties[:hostname]}"]
          resolutions = []
          resolutions << Resolution.new(Weaver::IWDUtils::RESOLUTION_LABELS[:change_credentials], res_description, Weaver::Core::ResolutionTypes::SET_CREDENTIALS)
        elsif !e.message.nil? && e.message.index("404")
          msg = Weaver::IWDUtils::MESSAGES[:invalid_pattern][:label] % ["#{pattern.name}", "#{pattern_id}", "#{iwd_properties[:hostname]}", "#{iwd_properties[:username]}"]
        sev_type = (is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:invalid_pattern][sev_type]
#          msg = msg + ", " + e.message
          res_description = Weaver::IWDUtils::MESSAGES[:import_another_pattern][:label] % ["#{pattern.name}", "#{pattern_id}"]
          resolutions = []
          resolutions << Resolution.new(Weaver::IWDUtils::RESOLUTION_LABELS[:import_another_pattern], res_description, Weaver::Core::ResolutionTypes::IMPORT_ANOTHER_PATTERN)
        else 
          severity = !other_severity.nil? ? other_severity : Logger::ERROR
          msg = !other_message.nil? ? e.to_s + other_message : e.to_s
          # TODO: add resolutions for other exceptions - maybe - set properties
          resolutions = []
        end
        problem_status = reporter.report(severity, msg, pattern) unless reporter.nil?
        problem_status.resolutions = resolutions
        @@logger.debug(msg)
      end
      
      def self.evaluated_resolutions(res_arr)
        evaluated = []
        res_arr.each { | resolution |
          begin
            label = resolution.label
            label = eval( '"' + resolution.label + '"') unless resolution.label.nil?
          rescue Exception => e         # exception is thrown when evaluating the label
            label = "Resolution label eval threw exception #{e}"
            @@logger.debug "rescued #{e}"
          end
          begin
            description = resolution.description
            description = eval( '"' + resolution.description + '"') unless resolution.description.nil?
          rescue Exception => e         # exception is thrown when evaluating the description
            description = "Resolution description eval threw exception #{e}"
            @@logger.debug "rescued #{e}"
          end
  
          evaluated << Resolution.new(label, description, resolution.type, resolution.fix_proc)
        }
        evaluated
      end
      
      # Print mutable attributes of a base object and its children to stdout
      #
      # * *Args*    :
      #   - +base+ -> the base object
      #   - +write_unnamed+ -> write unnamed attributes, defaults to false
      #
      def self.print_mutable_attributes(base, write_unnamed=false, indent = 0)
        indent_str = ""
        for i in 0..indent
          indent_str = indent_str + "\t"
        end
        puts "#{indent_str}#{base.id.to_s}:"
        base.amd.each do |id, attr|
          if(attr.mutable && !(attr.value.kind_of?(LateBinding)))
            if (write_unnamed || (!write_unnamed && !attr.name.nil?))
              #            puts(indent_str + "ATTRIBUTE #{attr.id}, #{attr.name.nil? ? "nil name" : attr.name}, mutable=#{attr.mutable}, value=#{attr.value}, valueClass=#{attr.value.class}, default=#{attr.default}, defaultClass=#{attr.default.class}")
              puts(indent_str + attr.to_s)
            end
          end
        end
        base.child_entities.each_value do |child|
          if (child.kind_of?(Base) && !child.kind_of?(Attribute))
            indent = indent + 1
            print_mutable_attributes(child, write_unnamed, indent)
            indent = indent - 1
          end
        end
      end

    end
  end
end