require 'fileutils'
require 'yaml'
require 'weaver/core/logging'
require 'weaver/core/weaver_error'
require 'weaver/system'
require 'pathname'
require 'hash' # Weaver's version, provides symbolize()

module Weaver
  module Utils
    # This class has helper methods for getting information about a Weaver project
    # Objects of this class are initialized with the name of the Weaver project folder
    class ProjectHelper
      include Weaver::Core

      WEAVER_FILE = '.weaver'
      DEFAULT_CONFIG_FILE = 'config.yml'
      # DEPRECATED
      LEGACY_DEPLOYMENT_PARAMETERS = :deployment_parameters
      WEAVER_PARAMETERS = :weaver
      DEFAULT_ENVIRONMENT = :default_environment
      
      # @return [Hash] configuration parameters
      attr_reader :config_params
      
      # @return [String[]] problems with the configuration config.yml file
      attr_reader :validation_messages
      
      # Instantiates a project helper
      # * *Args*    :
      # - +project_folder+ -> the name of the Weaver project folder or any subfolder
      # @param [Hash<String, Object>] config_overrides where keys in dotted notation refer to nested Hashes
      # @param [String] config_filename_override nil or a .yml filename to use instead of the configuration .yml identified in the .weaver directory 
      def initialize(project_folder, config_overrides, config_filename_override = nil)
        @config_params = nil
        @validation_messages = []

        if config_filename_override != nil
          unless File.exists? config_filename_override
            @@logger.error "Configuration file '#{config_filename_override}' not found"
            raise Weaver::Core::WeaverError, "Configuration file '#{config_filename_override}' not found"
          end
          
          @config_file_name = File.basename(config_filename_override)
          @project_folder = File.dirname(File.expand_path(config_filename_override))
          @config_file = config_filename_override
        else
          # An explicit nil means to run projectless
          if project_folder.nil?
            @config_file = @config_file_name = @project_folder = '<none>'
            @config_params = {}
            add_system_properties
            return
          end
          
          unless project_folder.kind_of? String
            @@logger.error "Project directory must be a string, but it is:'#{project_folder.class}'"
            raise Weaver::Core::WeaverError, "Project directory must be a string, but it is:'#{project_folder.class}'"
          end
  
          unless File.exists? project_folder
            @@logger.error "Project directory does not exist:'#{project_folder}'"
            raise Weaver::Core::WeaverError, "Given directory does not exist: '#{project_folder}'"
          end
  
          unless File.directory? project_folder
            @@logger.error "Given directory exists but is not a valid directory: '#{project_folder}'"
            raise Weaver::Core::WeaverError, "Given directory exists but is not a directory: '#{project_folder}'"
          end
  
          @project_folder = find_project_root(project_folder) # find dir by walking up the tree
          if @project_folder.nil?
            msg = "Cannot find valid Weaver project directory: '#{project_folder}'"
            @@logger.error msg
            raise Weaver::Core::WeaverError, msg
          end
  
          @project_folder = File.expand_path(@project_folder)
          
          weaver_file = File.join(@project_folder, WEAVER_FILE)
  
          @config_file_name = File.read(weaver_file).strip
          @config_file = File.join(@project_folder, @config_file_name)
          
          unless File.exists? @config_file
            msg = "Project folder '#{@project_folder}' does not have config file '#{@config_file}'"
            @@logger.error msg
            raise Weaver::Core::WeaverError, msg
          end
        end
        
        begin
#            @config_params = YAML::load_file(@config_file)
          @config_params = process_overrides(@config_file, config_overrides, @validation_messages)
          @config_params = @config_params.symbolize
        rescue SyntaxError => se
          msg = "Configuration file in Weaver project has errors: '#{@config_file}': #{se.inspect}"
          @@logger.error(msg)
          raise Weaver::Core::WeaverError.new(msg, se)
        end

        add_system_properties
        
        @validation_messages << "Deprecated #{@config_file_name} key #{LEGACY_DEPLOYMENT_PARAMETERS}" if @config_params.key?(LEGACY_DEPLOYMENT_PARAMETERS)
      end

      # Returns a Hash of the deployment parameters specified in config file under 'weaver'
      # * *Args*    :
      # - +hash_to_merge+ -> a hash of deployment parameters that need to be merged with the one in the config file
      # * *Returns* :
      #   - a  Hash of deployment parameters
      def deployment_parameters(hash_to_merge = {})
        if !@config_params.nil?
          deploy_params = @config_params[WEAVER_PARAMETERS] || @config_params[LEGACY_DEPLOYMENT_PARAMETERS] || {}
          if deploy_params.nil?
            @@logger.warn "Deployment parameters are not specified under '#{WEAVER_PARAMETERS}' in configuration file '#{@project_folder}/#{@config_file_name}'. "
          end
          unless hash_to_merge.nil?
            deploy_params.merge!(hash_to_merge)
          end
          return deploy_params.symbolize
        else
          @@logger.warn "Configuration parameters are not loaded in Weaver project '#{@project_folder}'"
          return nil
        end
      end

      # Given a provider id it returns as has with the provider data.
      # * *Args*    :
      # - +provider_id+ -> the provider id either as (String or Symbol)
      # * *Returns* :
      #   - a Has with the provider data
      def provider_data(provider_id)
        if @config_params.include? provider_id
          @config_params[provider_id]
        elsif @config_params.include? provider_id.to_sym
          @config_params[provider_id.to_sym]
        else
          msg = "Cannot find data for provider '#{provider_id}' in '#{@config_file_name}'."
          @@logger.error msg
          raise Weaver::Core::WeaverError, msg
        end
      end

      # Returns the path of the default environment file specified in config.yml under 'default_environment'
      # * *Returns* :
      #   - a String with the path of the default environment file to use
      def default_environment()
        if !@config_params.nil?
          default_environment = @config_params[DEFAULT_ENVIRONMENT]
          if default_environment.nil?
            @@logger.warn "Default_environment file is not specified under '#{DEFAULT_ENVIRONMENT}' in configuration file '#{@project_folder}/#{@config_file_name}'. "
          end
          environment_file = "#{@project_folder}/#{default_environment}"
          unless File.exists? environment_file
            @@logger.warn "Default_environment file does not exist: '#{environment_file}' "
          end
          return environment_file
        else
          @@logger.warn "Configuration parameters are not loaded in Weaver project '#{@project_folder}'"
          return nil
        end
      end

      protected
      
      # Read config_filename as a YAML file, override values using config_override, report errors via out_invalid_overrides
      def process_overrides(config_filename, config_overrides, out_invalid_overrides)
        config = {}
        unless config_overrides.nil?
          begin
            config = YAML.load_file(config_filename)
          rescue StandardError => se
            out_invalid_overrides << "Couldn't read configuration #{config_filename} #{se.to_s}"
          end
        end
        
        config_overrides.each do |key, value|
          keys = key.split('.')
          prop = keys.pop
          hsh = config
          pos = 0
          keys.each do |subkey|
            unless hsh.nil?
              hsh = hsh[subkey]
              if hsh.nil?
                out_invalid_overrides << "Unknown config key #{keys[0..pos].join('.')}"
              end
              pos = pos + 1
            end
          end
          
          unless hsh.nil?
            if hsh.key?(prop)
              propclass = hsh[prop].class
              if propclass == String
                value = value.to_s
              elsif propclass == Fixnum
                value = value.to_i
              elsif propclass == TrueClass || propclass == FalseClass
                value = value.to_b
              end
              hsh[prop] = value
            else
              out_invalid_overrides << "Unknown config key #{key}"
            end
          end
        end
        
        return config
      end

      # Add all system setting as Weaver system properties
      def add_system_properties()
        if @config_params.include? :system_settings
          @config_params[:system_settings].each do |k,v|
            Weaver::System.instance.set_property(k, v)
          end
        end
      end

      # Recursively tries to find the root project folder (the folder containing '.weaver')
      def find_project_root(folder)
        weaver_file = File.join(folder, WEAVER_FILE)
        return folder if File.exists?(weaver_file)

        folder = Pathname.new(File.expand_path("..", folder))
        return nil if folder.root?
        return find_project_root(folder)
      end

    end
  end

end
